"""Cache tables in postgres for the SPW dashboard."""
import xarray as xr
import numpy as np
from functools import partial
from nuthatch import cache
from sheerwater.utils import dask_remote, start_remote
from sheerwater.reanalysis import era5_rolled
from sheerwater.forecasts.ecmwf_er import ecmwf_ifs_spw
from sheerwater.forecasts.fuxi import fuxi_spw
from sheerwater.climatology import climatology_spw, climatology_rolled
from sheerwater.data.imerg import imerg_rolled
from sheerwater.data.chirps import chirps_rolled
from sheerwater.data.ghcn import _ghcn_rolled_unified
from sheerwater.data.tahmo import tahmo_rolled
from sheerwater.metrics import get_datasource_fn
from sheerwater.tasks import spw_precip_preprocess


@dask_remote
@cache(cache_args=['truth', 'grid', 'mask', 'region'],
       backend='sql')
def rain_windowed_spw(start_time, end_time,
                      truth='era5',
                      grid='global1_5', mask='lsm', region='global'):
    """Store the rolling windows of precipitation relevant to SPW in the database."""
    # Get the ground truth data
    if truth == 'era5':
        fn = partial(era5_rolled, start_time, end_time, variable='precip', grid=grid)
    elif truth == 'imerg':
        fn = partial(imerg_rolled, start_time, end_time, grid=grid)
    elif truth == 'chirps':
        fn = partial(chirps_rolled, start_time, end_time, grid=grid)
    elif truth == 'ghcn':
        fn = partial(_ghcn_rolled_unified, start_time, end_time, variable='precip', grid=grid,
                     missing_thresh=0.0, cell_aggregation='first')
    elif truth == 'tahmo':
        fn = partial(tahmo_rolled, start_time, end_time,
                     grid=grid, missing_thresh=0.0, cell_aggregation='first')
    elif truth == 'ltn':
        fn = partial(climatology_rolled, start_time, end_time, variable='precip',
                     first_year=2004, last_year=2015,
                     prob_type='deterministic', grid=grid)
    ds = spw_precip_preprocess(fn, mask=mask, region=region, grid=grid,
                               agg_days=[1, 8, 11], shift_days=[0, 0, 0])
    ds = ds.drop_vars('spatial_ref')
    df = ds.to_dataframe().dropna()
    return df


@dask_remote
@cache(backend='sql',
       cache_args=['truth', 'variable', 'grid', 'mask', 'region'])
def ea_rainy_onset_truth(start_time, end_time,
                         truth='era5',
                         variable='rainy_onset',
                         grid='global1_5', mask='lsm', region='global'):
    """Store the East African rainy season onset from a given truth source, according to the SPW method."""
    fn = get_datasource_fn(truth)
    kwargs = {'missing_thresh': 0.0} if truth in ['tahmo', 'ghcn'] else {}
    ds = fn(start_time, end_time, variable, agg_days=None,
            grid=grid, mask=mask, region=region, **kwargs)
    if 'spatial_ref' in ds.coords:
        ds = ds.drop_vars('spatial_ref')
    df = ds.to_dataframe()
    df = df.dropna(subset=variable)
    return df


@dask_remote
@cache(backend='sql',
       cache_args=['forecast', 'variable', 'grid', 'mask', 'region'])
def ea_rainy_onset_forecast(start_time, end_time,
                            forecast='ecmwf_ifs_er_debiased',
                            variable='rainy_onset',
                            grid='global1_5', mask='lsm', region='global'):
    """Store the rainy season onset from a given forecast source, according to the SPW method."""
    fn = get_datasource_fn(forecast)
    if forecast in ['ecmwf_ifs_er_debiased', 'ecmwf_ifs_er', 'fuxi']:
        if variable == 'rainy_onset_no_drought':  # 20 days instead of 21 to enable FuXi, which only goes out to 42 days
            leads = [f'day{d}' for d in [1, 7, 14, 20]]
        else:
            leads = [f'day{d}' for d in [1, 7, 14, 20, 28]]
    else:
        leads = ['day1']

    datasets = []
    for lead in leads:
        data = fn(start_time, end_time, variable, lead=lead,
                  grid=grid, mask=mask, region=region)
        datasets.append(data)

    ds = xr.concat(datasets, dim='lead_time')
    # if forecast == 'climatology_2015':
    # ds = ds.assign_coords(lead_time=[np.timedelta64(0, 'D')])
    df = ds.to_dataframe()
    df = df.dropna(subset=variable)
    return df


@dask_remote
@cache(backend='sql',
       cache_args=['forecast', 'variable', 'grid', 'mask', 'region'])
def ea_rainy_onset_probabilities(start_time, end_time,
                                 forecast='ecmwf_ifs_er_debiased',
                                 variable='rainy_onset',
                                 grid='global1_5', mask='lsm', region='global'):
    """Store the rainy season onset probabilities from a given forecast source, according to the SPW method."""
    datasets = []
    drought_condition = (variable == 'rainy_onset_no_drought')
    if forecast in ['ecmwf_ifs_er_debiased', 'ecmwf_ifs_er', 'fuxi']:
        if drought_condition:  # 20 days instead of 21 to enable FuXi, which only goes out to 42 days
            leads = [f'day{d}' for d in [1, 7, 14, 20]]
        else:
            leads = [f'day{d}' for d in [1, 7, 14, 20, 28]]
    else:
        leads = ['day1']

    for lead in leads:
        if forecast in ['ecmwf_ifs_er_debiased', 'ecmwf_ifs_er']:
            ds = ecmwf_ifs_spw(start_time, end_time, lead=lead,
                               debiased=(forecast == 'ecmwf_ifs_er_debiased'),
                               onset_group=None, drought_condition=drought_condition,
                               prob_type='probabilistic', prob_threshold=None,
                               grid=grid, mask=mask, region=region)
        elif forecast == 'fuxi':
            ds = fuxi_spw(start_time, end_time, lead=lead,
                          onset_group=None, drought_condition=drought_condition,
                          prob_type='probabilistic', prob_threshold=None,
                          grid=grid, mask=mask, region=region)
        elif forecast == 'climatology_2015':
            ds = climatology_spw(start_time, end_time, first_year=2004, last_year=2015,
                                 onset_group=None, drought_condition=drought_condition,
                                 prob_type='probabilistic', prob_threshold=None,
                                 grid=grid, mask=mask, region=region)
        else:
            raise ValueError(f"Invalid forecast source: {forecast}")
        datasets.append(ds)
    ds = xr.concat(datasets, dim='lead_time')
    if forecast == 'climatology_2015':
        ds = ds.assign_coords(lead_time=[np.timedelta64(0, 'D')])
    df = ds.to_dataframe()
    df = df.dropna(subset=variable)
    return df


if __name__ == "__main__":
    start_remote(remote_config='xlarge_cluster')
    # Runners to generate the tables
    start_time = '2016-01-01'
    end_time = '2022-12-31'
    grid = 'global1_5'
    mask = 'lsm'
    region = 'kenya'
    variables = ['rainy_onset_no_drought', 'rainy_onset']

    for variable in variables:
        for forecast in ['ecmwf_ifs_er_debiased', 'ecmwf_ifs_er', 'climatology_2015', 'fuxi']:
            df = ea_rainy_onset_forecast(start_time, end_time, forecast,
                                         variable=variable,
                                         grid=grid, mask=mask, region=region,
                                         #  recompute=True, force_overwrite=True,
                                         backend='sql')

            df = ea_rainy_onset_probabilities(start_time, end_time, forecast,
                                              variable=variable,
                                              grid=grid, mask=mask, region=region,
                                              #   recompute=True, force_overwrite=True,
                                              backend='sql')

        # Generate for all data sources
        for truth in ["era5", "chirps", "ghcn", "imerg", "tahmo"]:
            df2 = ea_rainy_onset_truth(start_time, end_time, truth,
                                       variable=variable,
                                       grid=grid, mask=mask, region=region,
                                       #    recompute=True,
                                       #    force_overwrite=True,
                                       backend='sql')
            # df3 = rain_windowed_spw(start_time, end_time, truth,
            # grid=grid, mask=mask, region=region,
            # backend='sql', recompute=True, force_overwrite=True)



"""Time and date utility functions for all parts of the data pipeline."""
from calendar import isleap
from datetime import datetime, timedelta

import dateparser
import numpy as np
import pandas as pd
import xarray as xr
from dateutil.rrule import DAILY, MONTHLY, WEEKLY, YEARLY, rrule

DATETIME_FORMAT = "%Y-%m-%d"


def shift_by_days(date, days):
    """Shifts a date by a given number of days."""
    input_type = type(date)

    # Convert input time to datetime object  for relative delta
    if isinstance(date, str):
        date_obj = dateparser.parse(date)
    elif isinstance(date, np.datetime64):
        date_obj = pd.Timestamp(date)
    elif isinstance(date, datetime):
        date_obj = date
    else:
        raise ValueError(f"Date type {type(date)} not supported.")

    # Shift the date
    new_date = date_obj + np.timedelta64(days, 'D')

    # Convert back to original type
    if np.issubdtype(input_type, str):
        new_date = datetime.strftime(new_date, "%Y-%m-%d")
    elif np.issubdtype(input_type, np.datetime64):
        new_date = np.datetime64(new_date, 'ns')
    return new_date


def dayofyear_to_datetime(x):
    """Converts day of year to datetime."""
    if np.isnan(x):
        return np.datetime64('NaT', 'ns')
    return np.datetime64("1904-01-01", 'ns') + np.timedelta64(int(x), 'D')


def assign_grouping_coordinates(ds, group, time_dim='time'):
    """Assigns time-based grouping coordinates to the time dimension.

    Args:
        ds (xr.Dataset): The dataset to add grouping to.
        group (str, list): The time grouping label to add. One of:
            - 'month': Group by month.
            - 'quarter': Group by quarter.
            - 'ea_rainy_season': Group by East African rainy season.
            - 'year': Group by year.
        time_dim (str): The time dimension to add the grouping based on.
    """
    # Ensure grouping is in standard format
    if group and not isinstance(group, list):
        group = [group]
    coords = []
    for grp in group:
        if grp == 'month':
            coords.append(ds[time_dim].dt.month.values)
        elif grp == 'quarter':
            coords.append(ds[time_dim].dt.quarter.values)
        elif grp == 'ea_rainy_season':
            # East African rainy period is from March to May and October to December
            def month_to_period(month):
                if 2 <= month <= 5:
                    return 'MAM'
                elif 9 <= month <= 12:
                    return 'OND'
                else:
                    return None
            coords.append([month_to_period(x) for x in ds[time_dim].dt.month.values])
        elif grp == 'year':
            coords.append(ds[time_dim].dt.year.values)
        else:
            raise ValueError(f"Invalid time grouping {grp}")

    def group_to_str(coord):
        coord_str = []  # string representation of the group
        for y in coord:  # iterate over all group values
            if y is None:  # if any values are None, give a none grouping value
                return None
            try:
                coord_str.append(f"{y:02d}")  # if the group value is a number, format it as a string
            except ValueError:  # MAM / OND strings
                coord_str.append(y)
        return "-".join(coord_str)  # join the group values with a dash

    joined_coords = [group_to_str(x) for x in zip(*coords)]
    ds = ds.assign_coords(group=(ds[time_dim].dims, joined_coords))
    return ds


def convert_group_to_time(group, groupby):
    """Converts a group label to a time coordinate.

    Assumes that groups are represented in the following format:
        - Years are a the year value
        - Months are 'M1', 'M2', ..., 'M12'
        - Quarters are 'Q1', 'Q2', ..., 'Q4'
        - East African rainy season is 'MAM', 'OND'

    Converts groups to datetime objects as follows:
        - Years are represented as the first day of the year
        - Months are represented as the first day of that month
        - Quarters are represented as the first day of the quarter
            (e.g., Q1 is Jan 1st, Q2 is Apr 1st, Q3 is Jul 1st, Q4 is Oct 1st)
        - East African rainy season is represented as the first day of the rainy season
            (e.g., the MAM rains are Mar 1st, the OND rains are Oct 1st)
    """
    def convert_to_datetime(entry, grouping):
        # Default values
        yy = '1904'
        mm = '01'
        dd = '01'
        if entry is None:
            return None
        for grp, val in zip(grouping, entry.split('-')):
            if grp == 'month':
                mm = f"{int(val):02d}"
            elif grp == 'quarter':
                mm = f"{(int(val)-1)*3+1:02d}"
            elif grp == 'ea_rainy_season':
                if val == 'MAM':
                    mm = '03'  # March rains
                elif val == 'OND':
                    mm = '10'  # October rains
                else:
                    raise ValueError(f"Invalid East African rainy season {val}")
            elif grp == 'year':
                yy = val
            else:
                raise ValueError(f"Invalid time grouping {grp}")
        return np.datetime64(f"{yy}-{mm}-{dd}", 'ns')
    # Ensure grouping is in standard format
    if groupby and not isinstance(groupby, list):
        groupby = [groupby]
    return [convert_to_datetime(x, groupby) for x in group.values]


def date_mean(data, dim='time'):
    """Mean of a date variables over a dimension, using integer number of days since 1970-01-01."""
    #  Get the number of days since 1970-01-01
    days = (data - np.datetime64('1970-01-01')).astype('timedelta64[D]').astype(int)
    days = days.where(days > -1000, np.nan)  # handle NaT, which is cast to a huge negative number
    days = days.mean(dim=dim, skipna=True)
    # Convert back to datetime, using integer number of days since 1970-01-01
    days = days.astype('datetime64[D]')
    return days


def doy_mean(data, dim='time'):
    """Mean of a date variables over a dimension, using day of year."""
    #  Get the number of days since 1970-01-01
    days = data.dt.dayofyear
    days = days.mean(dim=dim, skipna=True)
    # Convert back to datetime in the year 1904
    days = days.astype('datetime64[D]')
    days = days + (np.datetime64('1904-01-01') - np.datetime64('1970-01-01'))
    return days


def groupby_time(ds, groupby, agg_fn, time_dim='time', return_timeseries=False, **kwargs):
    """Aggregates data in groups along the time dimension according to time_grouping.

    Args:
        ds (xr.Dataset): The dataset to aggregate.
        groupby (str, list): The time grouping to use. Should be a string, a list, or a list of lists.
            List values should be one of:
            - 'month': Group by month.
            - 'year': Group by year.
            - 'quarter': Group by quarter.
            - 'ea_rainy_season': Group by East African rainy season (MAM and OND).
        agg_fn (object, list): The aggregation function to apply.
        time_dim (str): The time dimension to group by.
        return_timeseries (bool): If True, return a timeseries (the first date in each period).
             Otherwise, returns the label for the group (e.g., 'MAM', 2015, 'Q4', 'M1').
        kwargs: Additional keyword arguments to pass to the aggregation function.
    """
    # Input validation: if agg_fn is not a list, convert to a list
    if isinstance(groupby, list) and not isinstance(agg_fn, list):
        agg_fn = [agg_fn] * len(groupby)
    elif not isinstance(groupby, list) and not isinstance(agg_fn, list):
        groupby = [groupby]
        agg_fn = [agg_fn]
    elif not isinstance(groupby, list) and isinstance(agg_fn, list):
        raise ValueError("Cannot apply multiple aggregation functions to a single group.")
    if len(groupby) != len(agg_fn):
        raise ValueError("Length of group_by and agg_fn must be the same.")

    # Run multiple grouping steps
    for grp, agg in zip(groupby, agg_fn):
        # If no grouping is specified, apply the aggregation function directly
        if grp is None:
            ds = agg(ds, **kwargs)
        else:
            ds = assign_grouping_coordinates(ds, grp, time_dim)
            ds = ds.groupby("group").map(agg, **kwargs)
            if 'group' not in ds.dims:
                raise ValueError("Aggregation function must compress dataset along the group dimension.")

        # Convert group to time
        if (return_timeseries or len(groupby) > 1) and 'group' in ds.coords:
            ds = ds.assign_coords(time=("group", convert_group_to_time(ds['group'], grp)))
            ds = ds.swap_dims({'group': 'time'})
            ds = ds.sortby('time')
            ds = ds.drop('group')

    return ds


def add_dayofyear(ds, time_dim="time"):
    """Add a day of year coordinate to time dim, in the year 1904 (a leap year)."""
    ds = ds.assign_coords(dayofyear=(
        time_dim, pd.to_datetime([dateparser.parse(f"1904-{m}-{d}") for d, m in
                                  zip(ds[time_dim].dt.day.values, ds[time_dim].dt.month.values)])))
    return ds


def pad_with_leapdays(ds, time_dim="time"):
    """Pad the dataset with pseudo leap days in every year.

    Requires the input dataframe to have a dayofyear column. Modifies the input dataset.
    """
    # Find the years that don't have a leap day
    missing_leaps = [x for x in np.unique(ds[time_dim].dt.year.values) if not isleap(x)]
    missing_dates = [dateparser.parse(f"{x}-02-28") for x in missing_leaps]

    # Get the value on the 28th for these years
    missing_ds = ds.sel({time_dim: missing_dates})

    # Set the day of year to the 29th for these years
    missing_ds['dayofyear'] = (time_dim, [dateparser.parse("1904-02-29")]*len(missing_dates))

    # Add these new dates into the dataset and sort
    ds = xr.concat([ds, missing_ds], dim=time_dim)
    ds = ds.sortby(time_dim)
    return ds


def generate_dates_in_between(start_time, end_time, date_frequency, return_string=False):
    """Generates dates between two dates based on the frequency.

    Args:
        start_time (str): The start date.
        end_time (str): The end date.
        date_frequency (str): The frequency of the dates.
            One of "daily", "weekly", a day of the week (e.g., "Monday"), or a combination of days
            separated by a slash (e.g., "Monday/Thursday").
        return_string (bool): Whether to return the dates as strings or datetime objects.
    """
    start_date = dateparser.parse(start_time)
    end_date = dateparser.parse(end_time)

    if date_frequency not in ["daily", "weekly"]:
        dates = [
            date
            for date in generate_dates_in_between(start_time, end_time, "daily",
                                                  return_string=False)
            if date.strftime("%A") in date_frequency.split("/")
        ]
    else:
        frequency_to_int = {"daily": 1, "weekly": 7}
        dates = [
            start_date +
            timedelta(days=x * frequency_to_int[date_frequency])
            for x in range(0, int((end_date - start_date).days /
                                  (frequency_to_int[date_frequency])) + 1,)
        ]

    if return_string:
        dates = [date.strftime(DATETIME_FORMAT) for date in dates]
    return dates


def is_valid_forecast_date(model, forecast_type, forecast_date):
    """Checks if the forecast date is valid for the given model and forecast type."""
    valid_forecast_dates = {
        "reforecast": {
            "ecmwf": ("2015-05-14", datetime.today().strftime(DATETIME_FORMAT), "Monday/Thursday"),
            "salient": ("2022-01-01", datetime.today().strftime(DATETIME_FORMAT), "Wednesday"),
        },
        "forecast": {
            "ecmwf": ("2015-05-14", datetime.today().strftime(DATETIME_FORMAT), "Monday/Thursday"),
            "salient": ("2022-01-01", datetime.today().strftime(DATETIME_FORMAT), "Wednesday"),
        },
    }
    assert isinstance(forecast_date, datetime)
    try:
        return forecast_date in generate_dates_in_between(
            *valid_forecast_dates[forecast_type][model], return_string=False)
    except KeyError:
        return False


def get_dates(start_time, end_time, stride="day", return_string=True):
    """Outputs the list of dates corresponding to input date string."""
    # Input is of the form '20170101-20180130'
    start_date = dateparser.parse(start_time)
    end_date = dateparser.parse(end_time)

    if stride == "day":
        stride = DAILY
    elif stride == "week":
        stride = WEEKLY
    elif stride == "month":
        stride = MONTHLY
    elif stride == "year":
        stride = YEARLY
    else:
        raise ValueError(
            "Only day, week, month, and year strides are supported.")
    dates = [dt for dt in rrule(stride, dtstart=start_date, until=end_date)]
    if return_string:
        dates = [date.strftime(DATETIME_FORMAT) for date in dates]
    return dates



"""Library of metrics implementations for verification."""
# flake8: noqa: D102

import numpy as np
import pandas as pd
import xarray as xr


def mean_or_sum(ds, agg_fn, dims=['lat', 'lon']):
    """A light wrapper around standard groupby aggregation functions."""
    # Note, for some reason:
    # ds.groupby('region').mean(['lat', 'lon'], skipna=True).compute()
    # raises:
    # *** AttributeError: 'bool' object has no attribute 'blockwise'
    # or
    # *** TypeError: reindex_intermediates() missing 1 required positional argument: 'array_type'
    # So we have to do it via apply
    if agg_fn == 'mean':
        return ds.mean(dims, skipna=True)
    else:
        return ds.sum(dims, skipna=True)


def groupby_time(ds, time_grouping, agg_fn='mean'):
    """Aggregate a statistic over time."""
    if time_grouping is not None:
        if time_grouping == 'month_of_year':
            coords = [f'M{x:02d}' for x in ds.time.dt.month.values]
        elif time_grouping == 'year':
            coords = [f'Y{x:04d}' for x in ds.time.dt.year.values]
        elif time_grouping == 'quarter_of_year':
            coords = [f'Q{x:02d}' for x in ds.time.dt.quarter.values]
        elif time_grouping == 'day_of_year':
            coords = [f'D{x:03d}' for x in ds.time.dt.dayofyear.values]
        elif time_grouping == 'month':
            coords = [f'{pd.to_datetime(x).year:04d}-{pd.to_datetime(x).month:02d}-01' for x in ds.time.values]
        else:
            raise ValueError("Invalid time grouping")
        ds = ds.assign_coords(group=("time", coords))

        if agg_fn == 'mean':
            ds = ds.groupby("group").mean(dim="time", skipna=True)
        else:
            ds = ds.groupby("group").sum(dim="time", skipna=True)
        ds = ds.rename({"group": "time"})
        ds = ds.assign_coords(time=ds['time'].astype('<U10'))
    else:
        # Average in time
        if agg_fn == 'mean':
            ds = ds.mean(dim="time")
        elif agg_fn == 'sum':
            ds = ds.sum(dim="time")
        else:
            raise ValueError(f"Invalid aggregation function {agg_fn}")
    return ds


def groupby_region(ds, region_ds, mask_ds, agg_fn='mean', weighted=False):
    """Aggregate a statistic over region."""
    if agg_fn not in ['mean', 'sum']:
        raise ValueError(f"Invalid aggregation function {agg_fn}")
    if weighted and agg_fn == 'sum':
        raise ValueError("Cannot aggregate by sum with weighted averaging")

    if 'region' not in ds.coords:
        ds = ds.assign_coords(region=region_ds.region)

    variable_names = list(ds.data_vars)

    # Aggregate in space
    if weighted:
        # Group by region and average in space, while applying weighting for mask
        weights = latitude_weights(ds, lat_dim='lat', lon_dim='lon')
        # Expand weights to have a time dimension that matches ds
        if 'time' in ds.dims:  # Enable a time specific null pattern
            weights = weights.expand_dims(time=ds.time)

        # Ensure the weights null pattern matches the ds null pattern
        # Get all variable names in the dataset (excluding coords)
        weights = weights.where(ds[variable_names[0]].notnull(), np.nan, drop=False)

        # Mulitply by weights
        weights = weights * mask_ds.mask
    else:
        weights = mask_ds.mask

    ds['weights'] = weights
    for var in variable_names:
        ds[var] = ds[var] * ds['weights']

    ds = ds.groupby('region').sum(dim=['lat', 'lon'], skipna=True)

    if agg_fn == 'mean':
        for var in variable_names:
            ds[var] = ds[var] / ds['weights']

    ds = ds.drop_vars(['weights'])
    return ds


def latitude_weights(ds, lat_dim='lat', lon_dim='lon'):
    """Return latitude weights as an xarray DataArray.

    This function weights each latitude band by the actual cell area,
    which accounts for the fact that grid cells near the poles are smaller
    in area than those near the equator.
    """
    # Calculate latitude cell bounds
    lat_rad = np.deg2rad(ds[lat_dim].values)
    pi_over_2 = np.array([np.pi / 2], dtype=ds[lat_dim].dtype)
    if lat_rad.min() == -pi_over_2 and lat_rad.max() == pi_over_2:
        # Dealing with the full globe
        lower_bound = -pi_over_2
        upper_bound = pi_over_2
    else:
        # Compute the difference in between cells
        diff = np.diff(lat_rad)
        if diff.std() > 0.5:
            raise ValueError(
                "Nonuniform grid! Need to think about spatial averaging more carefully.")
        diff = diff.mean()
        lower_bound = np.array([lat_rad[0] - diff / 2.0], dtype=lat_rad.dtype)
        upper_bound = np.array([lat_rad[-1] + diff / 2.0], dtype=lat_rad.dtype)

    bounds = np.concatenate([lower_bound, (lat_rad[:-1] + lat_rad[1:]) / 2, upper_bound])

    # Calculate cell areas from latitude bounds
    upper = bounds[1:]
    lower = bounds[:-1]
    # normalized cell area: integral from lower to upper of cos(latitude)
    weights = np.sin(upper) - np.sin(lower)

    # Normalize weights
    weights /= np.mean(weights)
    # Return an xarray DataArray with dimensions lat
    weights = xr.DataArray(weights, coords=[ds[lat_dim]], dims=[lat_dim])
    weights = weights.expand_dims({lon_dim: ds[lon_dim]})
    return weights



"""Fetches ERA5 data from the Google ARCO Store.

NOTE: ERA5 land implementation is in progress.
"""
import numpy as np
import xarray as xr
from dateutil import parser
from nuthatch import cache
from nuthatch.processors import timeseries

from sheerwater.interfaces import data as sheerwater_data, spatial

from sheerwater.utils import (
    dask_remote,
    get_grid,
    get_grid_ds,
    get_variable,
    lon_base_change,
    regrid,
    roll_and_agg,
)
from sheerwater.utils.secrets import earth_data_hub_token


@dask_remote
@timeseries()
@spatial()
@cache(cache=False,
       cache_args=['variable'])
def era5_land_raw(start_time, end_time, variable, mask=None, region='global'):  # noqa ARG001
    """ERA5 function that returns data from earth data hub."""
    token = earth_data_hub_token()
    ds = xr.open_dataset(
        f"https://edh:{token}@data.earthdatahub.destine.eu/era5/reanalysis-era5-land-no-antartica-v0.zarr",
        chunks={},
        # chunks={'time': 365, 'latitude': 300, 'longitude': 300},
        engine="zarr",
    )

    # Select the right variable
    if variable in ['tmax2m', 'tmin2m']:
        var = 'tmp2m'  # Compute min and max daily temperatures from 2m temperature
    var = get_variable(variable, 'era5_land')
    ds = ds[var].to_dataset()

    # Convert local dataset naming and units
    ds = ds.rename({'latitude': 'lat', 'longitude': 'lon', 'valid_time': 'time'})

    # Raw latitudes are in descending order
    ds = ds.sortby('lat')
    ds = ds.rename_vars(name_dict={var: variable})

    return ds


@dask_remote
@spatial()
@cache(cache_args=['year', 'variable'],
       backend_kwargs={
           'chunking': {"lat": 300, "lon": 300, "time": 365}
})
def era5_land_daily_year(year, variable, mask=None, region='global'):
    """Aggregates the hourly ERA5 data into daily data.

    Args:
        year: The year to fetch data for.
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
            - global0_25: 0.25 degree global grid
        mask (str): The mask to apply to the data.
        region (str): The region to clip the data to.
    """
    # Read and combine all the data into an array
    start_time = str(year) + '-01-01'
    end_time = str(year) + '-12-31'
    ds = era5_land_raw(start_time, end_time, variable, mask=mask, region=region)

    # Convert to base180 longitude
    ds = lon_base_change(ds, to_base="base180")

    K_const = 273.15
    if variable == 'tmp2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(time='D').mean(dim='time')
    elif variable == 'tmax2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(time='D').max(dim='time')
    elif variable == 'tmin2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(time='D').min(dim='time')
    elif variable == 'precip':
        ds[variable] = ds[variable] * 1000.0
        ds.attrs.update(units='mm')
        ds = ds.resample(time='D').sum(dim='time')
        # Can't have precip less than zero (there are some very small negative values)
        ds = np.maximum(ds, 0)
    elif variable == 'ssrd':
        ds = ds.resample(time='D').sum(dim='time')
        ds = np.maximum(ds, 0)
    else:
        raise ValueError(f"Variable {variable} not implemented.")

    ds = ds.chunk({"lat": 300, "lon": 300, "time": 365})
    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable'],
       backend_kwargs={
           'chunking': {"lat": 300, "lon": 300, "time": 365},
})
def era5_land_daily(start_time, end_time, variable, mask=None, region='global'):
    """Aggregates the hourly ERA5 data into daily data for a date range.

    Args:
        start_time: Start date for the data range.
        end_time: End date for the data range.
        variable: Variable to fetch.
        mask (str): The mask to apply to the data.
        region (str): The region to clip the data to.
    """
    years = range(parser.parse(start_time).year, parser.parse(end_time).year + 1)

    datasets = []
    for year in years:
        ds = era5_land_daily_year(year, variable, mask=mask, region=region, filepath_only=True)
        datasets.append(ds)

    ds = xr.open_mfdataset(datasets,
                           engine='zarr',
                           parallel=True,
                           chunks={'lat': 300, 'lon': 300, 'time': 365})

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'grid'],
       cache_disable_if={'grid': 'global0_1'},
       backend_kwargs={
           'chunking': {"lat": 300, "lon": 300, "time": 365},
})
def era5_land_daily_regrid(start_time, end_time, variable, grid="global0_1", mask=None, region='global'):
    """ERA5 daily reanalysis with regridding."""
    ds = era5_land_daily(start_time, end_time, variable, mask=mask, region=region)
    _, _, size, _ = get_grid(grid)
    grid_ds = get_grid_ds(grid)

    ds = ds.reindex_like(grid_ds, method='nearest', tolerance=0.005)

    if grid == 'global0_1':
        return ds

    if size < 0.1:
        raise NotImplementedError("Unable to regrid ERA5 Land smaller than 0.1x0.1")
    else:
        # Regrid onto appropriate grid
        ds = regrid(ds, grid, base='base180', method='conservative', region=region)
    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'agg_days', 'grid'],
       cache_disable_if={'agg_days': 1},
       backend_kwargs={
           'chunking': {"lat": 300, "lon": 300, "time": 365},
})
def era5_land_rolled(start_time, end_time, variable, agg_days=7, grid="global0_1", mask=None, region='global'):
    """Aggregates the hourly ERA5 data into daily data and rolls.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        agg_days (int): The aggregation period, in days.
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
            - global0_25: 0.25 degree global grid
        mask (str): The mask to apply to the data.
        region (str): The region to clip the data to.
    """
    # Read and combine all the data into an array
    ds = era5_land_daily_regrid(start_time, end_time, variable, grid=grid, mask=mask, region=region)
    if agg_days == 1:
        return ds

    ds = roll_and_agg(ds, agg=agg_days, agg_col="time", agg_fn="mean")
    return ds


@dask_remote
@sheerwater_data()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def era5_land(start_time, end_time, variable, agg_days, grid='global0_1', mask='lsm', region='global'): # noqa: ARG001
    """Standard format task data for ERA5 Reanalysis.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        agg_days (int): The aggregation period, in days. Ignored if variable is 'rainy_onset'.
        grid (str): The grid resolution to fetch the data at.
        mask (str): The mask to apply to the data.
        region (str): The region to clip the data to.
    """
    _, _, size, _ = get_grid(grid)

    if size < 0.1 or (size == 0.1 and grid != 'global0_1'):
        raise NotImplementedError("Unable to regrid ERA5 Land smaller than 0.1x0.1")

    # Get daily data
    ds = era5_land_rolled(start_time, end_time, variable, agg_days=agg_days, grid=grid, mask=mask, region=region)
    return ds


@dask_remote
@spatial()
@timeseries()
@cache(cache=False,
       cache_args=['variable', 'grid'])
def era5_raw(start_time, end_time, variable, grid="global0_25", mask=None, region='global'):  # noqa ARG001
    """ERA5 function that returns data from Google ARCO."""
    if grid != 'global0_25':
        raise NotImplementedError(
                "Only ERA5 native 0.25 degree grid is implemented for the raw, hourly data."
                "Please call the era5_daily_regrid function to access daily ERA5 data on other grids.")

    # Pull the google dataset
    ds = xr.open_zarr('gs://gcp-public-data-arco-era5/ar/full_37-1h-0p25deg-chunk-1.zarr-v3',
                      chunks={'time': 50, 'latitude': 721, 'longitude': 1440})

    # Select the right variable
    if variable in ['tmax2m', 'tmin2m']:
        var = 'tmp2m'  # Compute min and max daily temperatures from 2m temperature
    var = get_variable(variable, 'era5')
    ds = ds[var].to_dataset()

    # Convert local dataset naming and units
    ds = ds.rename({'latitude': 'lat', 'longitude': 'lon'})

    # Raw latitudes are in descending order
    ds = ds.sortby('lat')
    ds = ds.rename_vars(name_dict={var: variable})

    return ds


@dask_remote
@spatial()
@timeseries()
@cache(cache_args=['variable', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 721, "lon": 1440, "time": 30}
})
def era5_daily(start_time, end_time, variable, grid="global1_5", mask=None, region='global'):
    """Aggregates the hourly ERA5 data into daily data.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
            - global0_25: 0.25 degree global grid
        mask (str): The mask to apply to the data.
        region (str): The region to clip the data to.
    """
    if grid != 'global0_25':
        raise ValueError("Only ERA5 native 0.25 degree grid is implemented.")

    # Read and combine all the data into an array
    ds = era5_raw(start_time, end_time, variable, grid='global0_25', mask=mask, region=region)

    # Convert to base180 longitude
    ds = lon_base_change(ds, to_base="base180")

    K_const = 273.15
    if variable == 'tmp2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(time='D').mean(dim='time')
    elif variable == 'tmax2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(time='D').max(dim='time')
    elif variable == 'tmin2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(time='D').min(dim='time')
    elif variable == 'precip':
        ds[variable] = ds[variable] * 1000.0
        ds.attrs.update(units='mm')
        ds = ds.resample(time='D').sum(dim='time')
        # Can't have precip less than zero (there are some very small negative values)
        ds = np.maximum(ds, 0)
    elif variable == 'ssrd':
        ds = ds.resample(time='D').sum(dim='time')
        ds = np.maximum(ds, 0)
    else:
        raise ValueError(f"Variable {variable} not implemented.")
    return ds


@dask_remote
@spatial()
@timeseries()
@cache(cache_args=['variable', 'grid'],
       cache_disable_if={'grid': 'global0_25'},
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "time": 1000},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'time': 30}
               }
           }
})
def era5_daily_regrid(start_time, end_time, variable, grid="global0_25", mask=None, region='global'):
    """ERA5 daily reanalysis with regridding."""
    ds = era5_daily(start_time, end_time, variable, grid='global0_25', mask=mask, region=region)
    ds = ds.sortby('lat')  # TODO: remove if we fix the era5 daily caches
    if grid == 'global0_25':
        return ds

    # Regrid onto appropriate grid
    # Need all lats / lons in a single chunk to be reasonable
    chunks = {'lat': 721, 'lon': 1440, 'time': 30}
    ds = ds.chunk(chunks)
    # Need all lats / lons in a single chunk for the output to be reasonable
    ds = regrid(ds, grid, base='base180', method='conservative', output_chunks={"lat": 121, "lon": 240}, region=region)
    return ds


@dask_remote
@spatial()
@timeseries()
@cache(cache_args=['variable', 'agg_days', 'grid'],
       cache_disable_if={'agg_days': 1},
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "time": 1000},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'time': 30}
               }
           }
})
def era5_rolled(start_time, end_time, variable, agg_days=7, grid="global1_5", mask=None, region='global'):
    """Aggregates the hourly ERA5 data into daily data and rolls.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        agg_days (int): The aggregation period, in days.
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
            - global0_25: 0.25 degree global grid
        mask (str): The mask to apply to the data.
        region (str): The region to clip the data to.
    """
    # Read and combine all the data into an array
    ds = era5_daily_regrid(start_time, end_time, variable, grid=grid, mask=mask, region=region)
    if agg_days == 1:
        return ds
    ds = roll_and_agg(ds, agg=agg_days, agg_col="time", agg_fn="mean")
    return ds


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def era5(start_time=None, end_time=None, variable='precip', agg_days=1, grid='global0_25', mask='lsm', region='global'): # noqa: ARG001
    """Standard format task data for ERA5 Reanalysis.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        agg_days (int): The aggregation period, in days. Ignored if variable is 'rainy_onset'.
        grid (str): The grid resolution to fetch the data at.
        mask (str): The mask to apply to the data.
        region (str): The region to clip the data to.
    """
    _, _, size, _ = get_grid(grid)
    if size < 0.25:
        raise NotImplementedError("Unable to regrid ERA5 smaller than 0.25x0.25")
    ds = era5_rolled(start_time, end_time, variable, agg_days=agg_days, grid=grid, mask=mask, region=region)
    return ds



"""Dashboard data for the Sheerwater Benchmarking project."""
from .metrics_tables import (weekly_metric_table, monthly_metric_table,
                             ground_truth_metric_table, biweekly_metric_table)

__all__ = ['weekly_metric_table', 'monthly_metric_table', 'ground_truth_metric_table', 'biweekly_metric_table']



"""Space and geography utility functions for all parts of the data pipeline."""
import numpy as np
import rioxarray  # noqa: F401 - needed to enable .rio attribute
import xarray as xr
import logging

from .region_utils import region_data

logger = logging.getLogger(__name__)


def get_globe_slice(ds, lon_slice, lat_slice, lon_dim='lon', lat_dim='lat', base="base180"):
    """Get a slice of the globe from the dataset.

    Handle the wrapping of the globe when slicing.

    Args:
        ds (xr.Dataset): Dataset to slice.
        lon_slice (np.ndarray): The longitude slice.
        lat_slice (np.ndarray): The latitude slice.
        lon_dim (str): The longitude column name.
        lat_dim (str): The latitude column name.
        base (str): The base of the longitudes. One of:
            - base180, base360
    """
    if base == "base360" and (lon_slice < 0.0).any():
        raise ValueError("Longitude slice not in base 360 format.")
    if base == "base180" and (lon_slice > 180.0).any():
        raise ValueError("Longitude slice not in base 180 format.")

    # Ensure that latitude is sorted before slicing
    ds = ds.sortby(lat_dim)

    wrapped = is_wrapped(lon_slice)
    if not wrapped:
        return ds.sel(**{lon_dim: slice(lon_slice[0], lon_slice[-1]),
                         lat_dim: slice(lat_slice[0], lat_slice[-1])})
    # A single wrapping discontinuity
    if base == "base360":
        slices = [[lon_slice[0], 360.0], [0.0, lon_slice[-1]]]
    else:
        slices = [[lon_slice[0], 180.0], [-180.0, lon_slice[-1]]]
    ds_subs = []
    for s in slices:
        ds_subs.append(ds.sel(**{
            lon_dim: slice(s[0], s[-1]),
            lat_dim: slice(lat_slice[0], lat_slice[-1])
        }))
    return xr.concat(ds_subs, dim=lon_dim)


def lon_base_change(ds, to_base="base180", lon_dim='lon'):
    """Change the base of the dataset from base 360 to base 180 or vice versa.

    Args:
        ds (xr.Dataset): Dataset to change.
        to_base (str): The base to change to. One of:
            - base180
            - base360
        lon_dim (str): The longitude column name.
    """
    if to_base == "base180":
        if (ds[lon_dim] < 0.0).any():
            print("Longitude already in base 180 format.")
            return ds
        lons = base360_to_base180(ds[lon_dim].values)
    elif to_base == "base360":
        if (ds[lon_dim] > 180.0).any():
            print("Longitude already in base 360 format.")
            return ds
        lons = base180_to_base360(ds[lon_dim].values)
    else:
        raise ValueError(f"Invalid base {to_base}.")

    # Check if original data is wrapped
    wrapped = is_wrapped(ds.lon.values)

    # Then assign new coordinates
    ds = ds.assign_coords({lon_dim: lons})

    # Sort the lons after conversion, unless the slice
    # you're considering wraps around the meridian
    # in the resultant base.
    if not wrapped:
        ds = ds.sortby('lon')
    return ds


def clip_region(ds, region, region_dim=None, region_df=None, lon_dim='lon', lat_dim='lat', drop=True):
    """Clip a dataset to a region.

    Args:
        ds (xr.Dataset): The dataset to clip to a specific region.
        region (str): The region to clip to. One of:
            - africa, conus, global
        region_dim (str): The name of the region dimension. If None, region data is fetched from the region registry.
        region_df (geopandas.GeoDataFrame): The region data to clip to. If None, is fetched from the region registry.
        lon_dim (str): The name of the longitude dimension.
        lat_dim (str): The name of the latitude dimension.
        drop (bool): Whether to drop the original coordinates that are NaN'd by clipping.
    """
    # No clipping needed
    if region == 'global':
        return ds

    if region_dim is not None:
        # If we already have a region dimension, just select the region
        ds = ds.sel(**{region_dim: region})
    else:
        # If we don't have a region dimension we need to clip to the region
        if region_df is None:
            # If not passed, fetch the region data from the region registry
            region_df = region_data(region)

        if len(region_df) != 1:
            raise ValueError(f"Region {region} has multiple geometries. Cannot clip.")
        # Set up dataframe for clipping
        ds = ds.rio.write_crs("EPSG:4326")
        ds = ds.rio.set_spatial_dims(lon_dim, lat_dim)
        # Clip the grid to the boundary of Shapefile
        ds = ds.rio.clip(region_df.geometry, region_df.crs, drop=drop)
    return ds


def apply_mask(ds, mask, var=None, val=0.0, grid='global1_5'):
    """Apply a mask to a dataset.

    Args:
        ds (xr.Dataset): Dataset to apply mask to.
        mask (str): The mask to apply. One of: 'lsm', None
        var (str): Variable to mask. If None, applies to apply to all variables.
        val (int): Value to mask below (any value that is
            strictly less than this value will be masked).
        grid (str): The grid resolution of the dataset.
    """
    # No masking needed
    if mask is None:
        return ds

    if isinstance(mask, str):
        from sheerwater.regions_and_masks import spatial_mask
        mask_ds = spatial_mask(mask, grid)
    else:
        mask_ds = mask

    # Check that the mask and dataset have the same dimensions
    if not all([dim in ds.dims for dim in mask_ds.dims]):
        raise ValueError("Mask and dataset must have the same dimensions.")

    if check_bases(ds, mask_ds) == -1:
        raise ValueError("Datasets have different longitude bases. Cannot mask.")

    # Ensure that the mask and the dataset don't have different precision
    # This MUST be np.float32 as of 4/28/25...unsure why?
    # Otherwise the mask doesn't match and lat/lons get dropped
    mask_ds['lon'] = np.round(mask_ds.lon, 5).astype(np.float32)
    mask_ds['lat'] = np.round(mask_ds.lat, 5).astype(np.float32)
    ds['lon'] = np.round(ds.lon, 5).astype(np.float32)
    ds['lat'] = np.round(ds.lat, 5).astype(np.float32)

    if isinstance(var, str):
        # Mask a single variable
        ds[var] = ds[var].where(mask_ds['mask'] > val, drop=False)
    else:
        # Mask multiple variables
        ds = ds.where(mask_ds['mask'] > val, drop=False)
    return ds


def snap_point_to_grid(point, grid_size, offset):
    """Snap a point to a provided grid and offset."""
    return round(float(point+offset)/grid_size) * grid_size - offset


def get_grid_ds(grid_id, base="base180", region='global'):
    """Get a dataset equal to ones for a given region."""
    lons, lats, _, _ = get_grid(grid_id, base=base)
    data = np.ones((len(lons), len(lats)))
    ds = xr.Dataset(
        {"mask": (['lon', 'lat'], data)},
        coords={"lon": lons, "lat": lats}
    )
    if region != 'global':
        ds = clip_region(ds, region=region)
    return ds


def get_grid(grid, base="base180"):
    """Get the longitudes, latitudes and grid size for a given global grid.

    Args:
        grid (str): The resolution to get the grid for. One of:
            - global1_5: 1.5 degree global grid
            - global0_25: 0.25 degree global grid
            - salient0_25: 0.25 degree Salient global grid
        base (str): The base grid to use. One of:
            - base360: 360 degree base longitude grid
            - base180: 180 degree base longitude grid
    """
    if grid == "global1_5":
        grid_size = 1.5
        offset = 0.0
    elif grid == "chirps":
        grid_size = 0.05
        offset = 0.025
    elif grid == "imerg":
        grid_size = 0.1
        offset = 0.05
    elif grid == "global0_25":
        grid_size = 0.25
        offset = 0.0
    elif grid == "global0_1":
        grid_size = 0.1
        offset = 0.0
    elif grid == "salient0_25":
        grid_size = 0.25
        offset = 0.125
    else:
        raise NotImplementedError(
            f"Grid {grid} has not been implemented.")

    # Instantiate the grid
    lons = np.arange(-180.0+offset, 180.0, grid_size)
    eps = 1e-6  # add a small epsilon to the end of the grid to enable poles for lat
    lats = np.arange(-90.0+offset, 90.0+eps, grid_size)
    if base == "base360":
        lons = base180_to_base360(lons)
        lons = np.sort(lons)

    # Round the longitudes and latitudes to the nearest 1e-5 to avoid floating point precision issues
    lons = np.round(lons, 5).astype(np.float32)
    lats = np.round(lats, 5).astype(np.float32)
    return lons, lats, grid_size, offset


def base360_to_base180(lons):
    """Converts a list of longitudes from base 360 to base 180.

    Args:
        lons (list, float): A list of longitudes, or a single longitude
    """
    if not isinstance(lons, np.ndarray) and not isinstance(lons, list):
        lons = [lons]
    val = [x - 360.0 if x >= 180.0 else x for x in lons]
    if len(val) == 1:
        return val[0]
    return np.array(val)


def base180_to_base360(lons):
    """Converts a list of longitudes from base 180 to base 360.

    Args:
        lons (list, float): A list of longitudes, or a single longitude
    """
    if not isinstance(lons, np.ndarray) and not isinstance(lons, list):
        lons = [lons]
    val = [x + 360.0 if x < 0.0 else x for x in lons]
    if len(val) == 1:
        return val[0]
    return np.array(val)


def is_wrapped(lons):
    """Check if the longitudes are wrapped.

    Works for both base180 and base360 longitudes. Requires that
    longitudes are in increasing order, outside of a wrap point.
    """
    wraps = (np.diff(lons) < 0.0).sum()
    if wraps > 1:
        raise ValueError("Only one wrapping discontinuity allowed.")
    elif wraps == 1:
        return True
    return False


def check_bases(ds, dsp, lon_col='lon', lon_colp='lon'):
    """Check if the bases of two datasets are the same."""
    if ds[lon_col].max() > 180.0:
        base = "base360"
    elif ds[lon_col].min() < 0.0:
        base = "base180"
    else:
        # Dataset base is ambiguous
        logger.info('Performing spatial data on a subset of data. Cannot check if longitude convention matches.'
                    'Please ensure that your dataframes are using the same longitude convention.')
        return 0

    if dsp[lon_colp].max() > 180.0:
        basep = "base360"
    elif dsp[lon_colp].min() < 0.0:
        basep = "base180"
    else:
        logger.info('Performing spatial data on a subset of data. Cannot check if longitude convention matches.'
                    'Please ensure that your dataframes are using the same longitude convention.')
        return 0

    # If bases are identifiable and unequal
    if base != basep:
        return -1
    return 0



"""The Suitable Planting Window (SPW) task utilities."""
from functools import partial

import xarray as xr

from sheerwater.utils import apply_mask, clip_region, doy_mean, first_satisfied_date, groupby_time


def rainy_onset_condition(da, prob_type='ensemble', prob_dim='member', prob_threshold=0.6):
    """Condition for the rainy season onset.

    Requires the input data to have 11d and 8d precipitation aggregations
    named 'precip_11d' and 'precip_8d' respectively.

    If prob_type is deterministic, return a boolean array.

    If a prob_type is probabilistic (ensemble, quantile) and prob_threshold is None,
    return the probability of the condition being met.

    If prob_type is probabilistic (ensemble, quantile) and prob_threshold is not None,
    return a boolean array based on the probability threshold.

    Args:
        da (xr.Dataset): Dataset to apply the condition to.
        prob_type (str): Type of probability to use, one of 'deterministic', 'ensemble', 'quantile'
        prob_dim (str): Name of the probabilistic dimension to average over
        prob_threshold (float): Threshold for the probability dimension.
    """
    # Ensure the data has the required aggregations
    if 'agg_precip_11d_shift0d' not in da.data_vars or 'agg_precip_8d_shift0d' not in da.data_vars:
        raise ValueError("Data must have 11d and 8d precipitation aggregations.")
    if prob_type in ['ensemble', 'quantile'] and prob_dim not in da.dims:
        raise ValueError("Probability dimension must be present for probabilistic forecasts.")

    # Check aggregation against the thresholds in mm
    cond = (da['agg_precip_11d_shift0d'] > 40.) & (da['agg_precip_8d_shift0d'] > 30.0)

    if prob_type == 'ensemble':
        # If the probability dimension is present
        cond = cond.mean(dim=prob_dim)
        if prob_threshold is not None:
            # Convert to a boolean based on the probability threshold
            cond = cond > prob_threshold
    elif prob_type == 'quantile':
        # If the quantile dimension is present
        raise NotImplementedError("Quantile probability not implemented yet.")
    return cond


def rainy_onset_no_drought_condition(da, prob_type='ensemble', prob_dim='member', prob_threshold=0.6):
    """Condition for the rainy season onset with a combined no drought condition.

    Requires the input data to have 11d and 8d precipitation aggregations
    named 'precip_11d', 'precip_8d', and 'precip_11d_shift11d' respectively.

    See `rainy_onset_condition` for more details.
    """
    # Ensure the data has the required aggregations
    if 'agg_precip_11d_shift0d' not in da.data_vars or 'agg_precip_8d_shift0d' not in da.data_vars \
            or 'agg_precip_11d_shift11d' not in da.data_vars:
        raise ValueError("Data must have 8d, 11d and 11d shifted 11d precipitation aggregations.")
    if prob_type in ['ensemble', 'quantile'] and prob_dim not in da.dims:
        raise ValueError("Probability dimension must be present for probabilistic forecasts.")

    # Check aggregation against the thresholds in mm
    cond = (da['agg_precip_11d_shift0d'] > 40.) & (
        da['agg_precip_8d_shift0d'] > 30.0) & (da['agg_precip_11d_shift11d'] > 15.)

    if prob_type == 'ensemble':
        # If the probability dimension is present
        cond = cond.mean(dim=prob_dim)
        if prob_threshold is not None:
            # Convert to a boolean based on the probability threshold
            cond = cond > prob_threshold
    elif prob_type == 'quantile':
        # If the quantile dimension is present
        raise NotImplementedError("Quantile probability not implemented yet.")
    return cond


def spw_precip_preprocess(agg_fn, shift_fn=None,
                          agg_days=None, shift_days=None,
                          mask='lsm', region='global', grid='global1_5', ):
    """Preprocess the daily data to ensure it has the required 8 and 11 day aggregations.

    Args:
        agg_fn (callable): Function that generates a n-day aggregated precipitation dataset when
            called with the 'agg_days' keyword argument.
        shift_fn (callable): Function that generates a n-day shifted precipitation dataset when
            called with the 'shift_days' keyword argument.
        agg_days (list): List of aggregation days to compute.
        shift_days (list): List of shift days to compute.
        mask (str): Mask to apply to the data.
        region (str): Region to clip the data to.
        grid (str): Grid to regrid the data to.
    """
    if agg_days is None:
        agg_days = [8, 11]
    if shift_days is None:
        shift_days = [0] * len(agg_days)
    # Default shift function to shift by the shift_days
    if shift_fn is None:
        def shift_fn(x, shift_by_days): return x.shift(time=-shift_by_days)

    # Ensure the data has the required aggregations
    try:
        datasets = [shift_fn(
            ad * agg_fn(agg_days=ad),
            shift_by_days=sd)
            .rename({'precip': f'agg_precip_{ad}d_shift{sd}d'})
            for ad, sd in zip(agg_days, shift_days)]
    except Exception as e:
        print("Aggregating and shifting failed. Check that you have defined a proper shifting function.")
        raise e

    # Remove lead_time dimension, which may not match across datasets
    if 'lead_time' in datasets[0]:
        datasets = [ds.drop('lead_time') for ds in datasets]

    # Merge both datasets
    ds = xr.merge(datasets)

    # Apply masking
    ds = apply_mask(ds, mask, grid=grid)
    ds = clip_region(ds, region=region)
    return ds


def spw_rainy_onset(ds, onset_group=None, aggregate_group=None, time_dim='time',
                    prob_type='ensemble', prob_dim=None, prob_threshold=None,
                    drought_condition=False,
                    mask='lsm', region='global', grid='global1_5'):
    """Utility function to get first rainy onset."""
    # Ensure that onset group and aggregate groups are lists
    if onset_group and not isinstance(onset_group, list):
        onset_group = [onset_group]
    if aggregate_group and not isinstance(aggregate_group, list):
        aggregate_group = [aggregate_group]

    # Form the groupby list for grouping utility
    groupby = [onset_group]
    if aggregate_group:
        groupby.append(aggregate_group)

    if drought_condition:
        condition = rainy_onset_no_drought_condition
    else:
        condition = rainy_onset_condition

    # If no grouping is provided, compute the condition probability on the timeseries
    if prob_threshold is None and prob_type in ['ensemble', 'quantile']:
        if onset_group is not None or aggregate_group is not None:
            raise ValueError("Grouping is not supported for probabilistic forecasts without a threshold.")
        # Compute condition probability on the timeseries
        rainy_da = condition(ds, prob_type=prob_type, prob_dim=prob_dim, prob_threshold=None)
    else:
        # Otherwise, compute the first satisfied date for each grouping
        agg_fn = [partial(first_satisfied_date,
                          condition=condition,
                          prob_type=prob_type, prob_dim=prob_dim, prob_threshold=prob_threshold,
                          time_dim=time_dim)]

        # Returns a dataframe with a dimension 'time' corresponding to the first grouping value
        # and value 'rainy_onset' corresponding to the rainy season onset date
        if aggregate_group is not None:
            # For the aggregate grouping average over datetimes within a grouping
            agg_fn += [partial(doy_mean, dim='time')]

        #  Apply the aggregation functions to the dataset using the groupby utility
        #  Set return_timeseries to True to get a dataset with a time dimension for each grouping
        rainy_da = groupby_time(ds, groupby=groupby, agg_fn=agg_fn,
                                time_dim=time_dim, return_timeseries=True)

    # Apply masking and clip to region
    var_name = 'rainy_onset' if not drought_condition else 'rainy_onset_no_drought'
    rainy_ds = rainy_da.to_dataset(name=var_name)
    rainy_ds = apply_mask(rainy_ds, mask, grid=grid)
    rainy_ds = clip_region(rainy_ds, region=region)
    return rainy_ds



"""Cache tables in postgres for the SPW dashboard."""
import xarray as xr

from nuthatch import cache
from sheerwater.utils import dask_remote, start_remote
from sheerwater.metrics import get_datasource_fn


@dask_remote
@cache(backend='sql',
       cache_args=['agg_days', 'grid', 'mask', 'region'])
def rainfall_data(start_time, end_time, agg_days=1,
                  grid='global1_5', mask='lsm', region='global'):
    """Store rainfall data across data sources to the database."""
    # Get the ground truth data
    datasets = []
    truth = ['era5', 'chirps', 'imerg', 'ghcn']
    if region in ['africa', 'east_africa']:
        truth.append('tahmo')
    for truth in truth:
        source_fn = get_datasource_fn(truth)
        ds = source_fn(start_time, end_time, 'precip', agg_days=agg_days,
                       grid=grid, mask=mask, region=region)
        ds = ds.rename({'precip': f'{truth}_precip'})
        datasets.append(ds)

    # Merge datasets
    ds = xr.merge(datasets, join='outer')
    ds = ds.drop_vars(['number', 'spatial_ref'])

    # Convert to dataframe
    df = ds.to_dataframe()
    return df


if __name__ == "__main__":
    start_remote()
    start_time = '2022-01-01'
    end_time = '2024-12-31'
    agg_days = [1, 7]
    for agg_day in agg_days:
        ds = rainfall_data(start_time, end_time, agg_day, grid='global0_25', mask='lsm', region='africa')



"""Test lead-based target date fetching."""
import numpy as np

from sheerwater.climatology import climatology_2015, climatology_agg_raw
from sheerwater.forecasts import salient
from sheerwater.forecasts.ecmwf_er import ifs_extended_range
from sheerwater.forecasts.salient import salient_blend
from sheerwater.reanalysis import era5, era5_rolled
from sheerwater.utils import convert_init_time_to_pred_time, shift_by_days


def test_target_date_conversion():
    """Test the conversion of target dates to forecast dates."""
    start_date = "2020-01-14"
    end_date = "2020-01-31"

    # Week2: shift back by 7 days
    fd_week2_start = shift_by_days(start_date, -7)
    fd_week2_end = shift_by_days(end_date, -7)
    assert fd_week2_start == "2020-01-07"
    assert fd_week2_end == "2020-01-24"

    # Weeks34: shift back by 14 days
    fd_week34_start = shift_by_days(start_date, -14)
    fd_week34_end = shift_by_days(end_date, -14)
    assert fd_week34_start == "2019-12-31"
    assert fd_week34_end == "2020-01-17"

    # Ground truth data is already in "target date" format
    ds = era5(start_date, end_date, "tmp2m", agg_days=7, grid="global1_5", mask=None, region='global')
    dsr = era5_rolled(start_date, end_date, "tmp2m", agg_days=7, grid="global1_5")
    assert ds.equals(dsr)

    # Climatology data is already in "target date" format
    ds = climatology_2015(start_date, end_date, "tmp2m", agg_days=7, grid="global1_5", mask=None, region='global')
    # Select week 2
    ds = ds.sel(prediction_timedelta=np.timedelta64(0, "D"), time="2020-01-14")
    dsr = climatology_agg_raw("tmp2m", 1985, 2014, agg_days=7, grid="global1_5")
    dsr = dsr.sel(dayofyear="1904-01-14")
    # Align coordinates for comparison
    dsr = dsr.rename({"dayofyear": "time"})
    dsr['time'] = ds['time']
    assert ds.equals(dsr)

    # Test forecast conversion
    ds = salient(start_date, end_date, "tmp2m", agg_days=7, grid="global1_5", mask=None, region='global')
    ds = ds.sel(prediction_timedelta=np.timedelta64(7, "D").astype("timedelta64[ns]"), time="2020-01-15")
    # Raw forecast
    dsr = salient_blend(fd_week2_start, fd_week2_end, "tmp2m", "sub-seasonal", grid="global1_5")
    dsr = dsr.sel(lead=2, quantiles=0.5, forecast_date="2020-01-08")
    dsr = dsr.rename({"forecast_date": "time"})
    dsr['time'] = ds['time']
    assert ((ds-dsr).tmp2m.max() < 1e-10)

    # Test forecast conversion using convert_init_time_to_pred_time
    ds = ifs_extended_range(fd_week2_start, fd_week2_end, "tmp2m", "forecast", grid="global1_5")
    # Get specific lead for week 2 (7 days)
    prediction_timedelta_shift = np.timedelta64(7, 'D').astype("timedelta64[ns]")
    ds = ds.sel(lead_time=prediction_timedelta_shift)

    # Calculate expected target times (init_time + prediction_timedelta)
    target_times = ds.start_date.values + ds.lead_time.values

    # Convert from init_time/prediction_timedelta to time using convert_init_time_to_pred_time
    # First rename to match expected dimensions
    ds = ds.rename({'start_date': 'init_time', 'lead_time': 'prediction_timedelta'})
    ds = ds.expand_dims({'prediction_timedelta': [prediction_timedelta_shift]})
    ds = convert_init_time_to_pred_time(ds)

    # Verify that the time coordinate matches the expected target times
    assert (ds.time.values == target_times).all()


if __name__ == "__main__":
    test_target_date_conversion()



"""Interface for gencast forecasts."""
import gcsfs
import numpy as np
import xarray as xr
from nuthatch import cache
from nuthatch.processors import timeseries

from sheerwater.utils import dask_remote, lon_base_change, regrid, roll_and_agg, shift_by_days
from sheerwater.interfaces import forecast as sheerwater_forecast, spatial


@dask_remote
@cache(cache_args=['year', 'variable', 'init_hour'],
       backend_kwargs={'chunking': {"lat": 721, "lon": 1440, 'lead_time': 10, 'time': 1, 'member': 5}})
def gencast_daily_year(year, variable, init_hour=0):
    """A daily Gencast forecast."""
    if init_hour != 0:
        raise ValueError("Only 0 init hour supported")

    # Get glob of all gencast forecasts
    fs = gcsfs.GCSFileSystem(project='sheerwater', token='google_default')

    def get_sub_ds(root):
        """Iterates through the root folder to find all sub zarrs for the requested year."""
        members = fs.ls(root, detail=False)
        members = members[1:]
        zarrs = []
        for mem in members:
            """Open a zarr for each member set."""
            mem_path = mem + '/forecasts_15d/'
            zarr = fs.ls(mem_path, detail=False)
            zarr = 'gs://' + zarr[0]
            ds = xr.open_zarr(zarr, chunks='auto', decode_timedelta=True)

            # Evaluation WIDs <= 7 are done with 0/12hr leads - >7 are 6/18hr leads.
            if ds.attrs['evaluation_wid'] <= 7:
                zarrs.append(zarr)
            else:
                print("Skipping 6/18hr leads.")

        return zarrs

    zarrs = []

    if year == '2020':
        zarrs += get_sub_ds('gs://weathernext/126478713_1_0/zarr/124883614_2020_to_2021')
    elif year == '2021':
        zarrs += get_sub_ds('gs://weathernext/126478713_1_0/zarr/124958964_2021_to_2022')
    elif year == '2022':
        zarrs += get_sub_ds('gs://weathernext/126478713_1_0/zarr/125057031_2022_to_2023')
    elif year == '2023':
        zarrs += get_sub_ds('gs://weathernext/126478713_1_0/zarr/125156073_2023_to_2024')

    ds = xr.open_mfdataset(zarrs,
                           chunks='auto',
                           decode_timedelta=True,
                           engine='zarr',
                           drop_variables=['100m_u_component_of_wind',
                                           '100m_v_component_of_wind',
                                           '10m_u_component_of_wind',
                                           '10m_v_component_of_wind',
                                           'geopotential',
                                           'mean_sea_level_pressure',
                                           'sea_surface_temperature',
                                           'specific_humidity',
                                           'temperature',
                                           'u_component_of_wind',
                                           'v_component_of_wind',
                                           'vertical_velocity'
                                           ])

    # Read the three years for gcloud
    ds = ds.rename({'prediction_timedelta': 'lead_time',
                    '2m_temperature': 'tmp2m',
                    'total_precipitation_12hr': 'precip',
                    'sample': 'member'})
    ds = ds[[variable]]

    # Select only the midnight initialization times
    ds = ds.where(ds.time.dt.hour == init_hour, drop=True)

    # Convert units
    K_const = 273.15
    if variable == 'tmp2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(lead_time='1D').mean(dim='lead_time')
    elif variable == 'precip':
        ds[variable] = ds[variable] * 1000.0
        ds.attrs.update(units='mm')

        # Convert from 12hrly to daily precip, robust to different numbers of 12hrly samples in a day
        ds = ds.resample(lead_time='1D').mean(dim='lead_time') * 2.0

        # Can't have precip less than zero (there are some very small negative values)
        ds = np.maximum(ds, 0)
    else:
        raise ValueError(f"Variable {variable} not implemented.")

    # Shift the lead back 12 hours + init time to be aligned
    ds['lead_time'] = ds['lead_time'] - np.timedelta64(init_hour+12, 'h')

    # Convert lat/lon
    ds = lon_base_change(ds)

    # This was necessary?
    ds = ds.chunk({"lat": 721, "lon": 1440, 'lead_time': 10, 'time': 1, 'member': 5})

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "lead_time": 10, "time": 10, 'member': 10},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'lead_time': 10, 'time': 1, 'member': 5}
               },
           }
})
def gencast_daily(start_time, end_time, variable, grid='global0_25', mask=None, region='global'):  # noqa: ARG001
    """A daily gencast forecast."""
    ds1 = gencast_daily_year(year='2020', variable=variable, init_hour=0)
    ds2 = gencast_daily_year(year='2021', variable=variable, init_hour=0)
    ds3 = gencast_daily_year(year='2022', variable=variable, init_hour=0)

    ds = xr.concat([ds1, ds2, ds3], dim='time')

    # This was necessary?
    ds = ds.chunk({'lat': 721, 'lon': 1440, 'lead_time': 10, 'time': 1, 'member': 5})

    # Regrid
    if grid != 'global0_25':
        ds = regrid(ds, grid, base='base180', method='conservative',
                    output_chunks={"lat": 721, "lon": 1440}, region=region)

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'agg_days', 'prob_type', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "lead_time": 10, "time": 10, "member": 10},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'lead_time': 10, 'time': 1, 'member': 5}
               },
           }
})
def gencast_rolled(start_time, end_time, variable, agg_days,
                   prob_type='deterministic', grid='global0_25', mask=None,
                   region='global'):
    """A rolled and aggregated gencast forecast."""
    ds = gencast_daily(start_time, end_time, variable, grid, mask=mask, region=region)

    if prob_type == 'deterministic':
        ds = ds.mean(dim='member')
        ds = ds.assign_attrs(prob_type="deterministic")
    else:
        ds = ds.assign_attrs(prob_type="ensemble")

    ds = roll_and_agg(ds, agg=agg_days, agg_col="lead_time", agg_fn="mean")
    return ds


@dask_remote
@sheerwater_forecast()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'prob_type', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365, 'lead_time': 1, 'member': 1}})
def gencast(start_time=None, end_time=None, variable="precip", agg_days=1, prob_type='deterministic',
            grid='global1_5', mask='lsm', region="global"):  # noqa: ARG001
    """Final Gencast interface."""
    if variable != 'precip':
        raise NotImplementedError("Data error present in non-precip variables in Gencast. Skipping.")

    # Get the data with the right days
    forecast_start = shift_by_days(start_time, -15) if start_time is not None else None
    forecast_end = shift_by_days(end_time, 15) if end_time is not None else None

    # Get the data with the right days
    ds = gencast_rolled(start_time=forecast_start, end_time=forecast_end, variable=variable,
                        agg_days=agg_days, prob_type=prob_type, grid=grid, mask=mask, region=region)
    if prob_type == 'deterministic':
        ds = ds.assign_attrs(prob_type="deterministic")
    else:
        ds = ds.assign_attrs(prob_type="ensemble")

    # Rename to standard naming
    ds = ds.rename({'time': 'init_time', 'lead_time': 'prediction_timedelta'})
    return ds



"""Utilities for running functions on a remote dask cluster."""
import logging
import os
import pwd
from functools import wraps

import coiled
from coiled.credentials.google import send_application_default_credentials
from dask.distributed import Client, LocalCluster, get_client

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

config_options = {
    'large_scheduler': {
        'scheduler_vm_types': ['c2-standard-16', 'c3-standard-22']
    },
    'xlarge_scheduler': {
        'scheduler_vm_types': ['c2-standard-30', 'c3-standard-44']
    },
    'xxlarge_scheduler': {
        'scheduler_vm_types': ['c3-standard-88']
    },
    'on_demand': {
        'spot_policy': 'on-demand'
    },
    'single_cluster': {
        'n_workers': 1
    },
    'large_cluster': {
        'n_workers': [10, 11]
    },
    'xlarge_cluster': {
        'n_workers': [15, 16]
    },
    'xxlarge_cluster': {
        'n_workers': [25, 26]
    },
    'xxxlarge_cluster': {
        'n_workers': [35, 36]
    },
    'xc_cluster': {
        'n_workers': [70, 71]
    },
    'cc_cluster': {
        'n_workers': [200, 201]
    },
    'large_node': {
        'worker_vm_types': ['c2-standard-16', 'c3-standard-22']
    },
    'xlarge_node': {
        'worker_vm_types': ['c2-standard-30', 'c3-standard-44']
    },
    'xxlarge_node': {
        'worker_vm_types': ['c3-standard-88']
    },
    'xxxlarge_node': {
        'worker_vm_types': ['c3-standard-176']
    },
    'large_disk': {
        'worker_disk_size': '150GiB'
    },
}


def start_remote(remote_name=None, remote_config=None):
    """Generic function to start a remote cluster."""
    default_name = 'sheerwater_' + pwd.getpwuid(os.getuid())[0]

    coiled_default_options = {
        'name': default_name,
        'n_workers': [3, 8],
        'idle_timeout': "120 minutes",
        'scheduler_vm_types': ['c2-standard-8', 'c3-standard-8'],
        'worker_vm_types': ['c2-standard-8', 'c3-standard-8'],
        'spot_policy': 'spot_with_fallback',
        'environ': {'NUTHATCH_ALLOW_INSTALLED_PACKAGE_CONFIGURATION': 'True'}
    }

    if remote_name and isinstance(remote_name, str):
        coiled_default_options['name'] = remote_name

    if remote_config and isinstance(remote_config, dict):
        # setup coiled cluster with remote config
        logger.info("Attaching to coiled cluster with custom configuration")
        coiled_default_options.update(remote_config)
    elif remote_config and (isinstance(remote_config, str) or
                            isinstance(remote_config, list)):
        logger.info("Attaching to coiled cluster with preset configuration")
        if not isinstance(remote_config, list):
            remote_config = [remote_config]

        for conf in remote_config:
            if conf in config_options:
                coiled_default_options.update(config_options[conf])
            else:
                print(f"Unknown preset remote config option {conf}. Skipping.")
    else:
        # Just setup a coiled cluster
        logger.info("Attaching to coiled cluster with default configuration")

    cluster = coiled.Cluster(**coiled_default_options)

    # send Application Default Credentials
    try:
        send_application_default_credentials(cluster)
    except Exception as e:
        print("Failed to send credentials", e)

    cluster.get_client()


def dask_remote(func):
    """Decorator to run a function on a remote dask cluster."""
    @wraps(func)
    def remote_wrapper(*args, **kwargs):
        # See if there are extra function args to run this remotely
        if 'remote' in kwargs and kwargs['remote']:

            remote_name = None
            if 'remote_name' in kwargs:
                remote_name = kwargs['remote_name']

            remote_config = None
            if 'remote_config' in kwargs:
                remote_config = kwargs['remote_config']

            start_remote(remote_name, remote_config)
        elif 'local_dask' in kwargs and kwargs['local_dask']:
            # Setup a local cluster
            try:
                get_client()
            except ValueError:
                logger.info("Starting local dask cluster...")
                cluster = LocalCluster(n_workers=2, threads_per_worker=2)
                Client(cluster)

        # call the function and return the result
        if 'remote' in kwargs:
            del kwargs['remote']

        if 'remote_config' in kwargs:
            del kwargs['remote_config']

        if 'remote_name' in kwargs:
            del kwargs['remote_name']

        if 'local_dask' in kwargs:
            del kwargs['local_dask']

        return func(*args, **kwargs)
    return remote_wrapper



#!/usr/bin/env python
"""Runs metrics and updates the caches."""
import itertools
import traceback

from sheerwater.utils import start_remote
from dashboard_data import biweekly_metric_table
from jobs import parse_args, run_in_parallel

(start_time, end_time, forecasts, metrics,
 variables, grids, regions, agg_days,
 time_groupings, parallelism, recompute,
 backend, remote_name, remote, remote_config) = parse_args()

if remote:
    start_remote(remote_config=remote_config, remote_name=remote_name)

combos = itertools.product(metrics, variables, grids, regions, time_groupings)

filepath_only = True
if backend is not None:
    filepath_only = False


def run_metrics_table(combo):
    """Run table metrics."""
    metric_name, variable, grid, region, time_grouping = combo

    try:
        biweekly_metric_table(start_time, end_time, variable, "era5", metric_name,
                              time_grouping=time_grouping, grid=grid, region=region,
                              cache_mode='overwrite', filepath_only=filepath_only,
                              recompute=recompute, storage_backend=backend)
    except KeyboardInterrupt as e:
        raise (e)
    except:  # noqa: E722
        print(f"Failed to run biweekly metric {grid} {variable} {metric_name} \
                {region} {time_grouping}: {traceback.format_exc()}")


if __name__ == "__main__":
    run_in_parallel(run_metrics_table, combos, parallelism)



"""Variable-related utility functions for all parts of the data pipeline."""
def convert_init_time_to_pred_time(ds, init_time_dim='init_time',
                                   lead_time_dim='prediction_timedelta', valid_time_dim='time'):
    """Convert the start_date and lead_time coordinates to a valid_time coordinate."""
    ds = ds.assign_coords({valid_time_dim: ds[init_time_dim] + ds[lead_time_dim]})
    tmp = ds.stack(z=(init_time_dim, lead_time_dim))
    tmp = tmp.set_index(z=(valid_time_dim, lead_time_dim))
    ds = tmp.unstack('z')
    ds = ds.rename({lead_time_dim: 'prediction_timedelta'})
    ds = ds.drop_vars(init_time_dim)
    return ds


def get_variable(variable_name, variable_type='era5'):
    """Converts a variable in any other type to a variable name of the requested type."""
    variable_ordering = ['sheerwater', 'era5', 'ecmwf_hres', 'ecmwf_ifs_er', 'salient', 'abc', 'ghcn', 'era5_land']

    weather_variables = [
        # Static variables (2):
        ('z', 'geopotential', 'geopotential', None, None, None, None, None),
        ('lsm', 'land_sea_mask', 'land_sea_mask', None, None, None, None, None),

        # Surface variables (6):
        ('tmp2m', '2m_temperature', '2m_temperature', '2m_temperature', 'temp', 'tmp2m', 'temp', 't2m'),
        ('precip', 'total_precipitation', 'total_precipitation_6hr', 'total_precipitation_24hr',
         'precip', 'precip', 'precip', 'tp'),
        ("vwind10m", "10m_v_component_of_wind", "10m_v_component_of_wind", None, None, None, None, 'v10'),
        ("uwind10m", "10m_u_component_of_wind", "10m_u_component_of_wind", None, None, None, None, 'u10'),
        ("msl", "mean_sea_level_pressure", "mean_sea_level_pressure", None, None, None, None, None),
        ("tisr", "toa_incident_solar_radiation", "toa_incident_solar_radiation", None, "tsi", None, None, 'ssr'),
        ("ssrd", "surface_solar_radiation_downwards", None, None, "tsi", None, None, 'ssrd'),

        # Atmospheric variables (6):
        ("tmp", "temperature", "temperature", None, None, None, None, None),
        ("uwind", "u_component_of_wind", "u_component_of_wind", None, None, None, None, None),
        ("vwind", "v_component_of_wind", "v_component_of_wind", None, None, None, None, None),
        ("hgt", "geopotential", "geopotential", None, None, None, None, None),
        ("q", "specific_humidity", "specific_humidity", None, None, None, None, None),
        ("w", "vertical_velocity", "vertical_velocity", None, None, None, None, None),
    ]

    name_index = variable_ordering.index(variable_type)

    for tup in weather_variables:
        for name in tup:
            if name == variable_name:
                val = tup[name_index]
                if val is None:
                    raise ValueError(f"Variable {variable_name} not implemented.")
                return val

    raise ValueError(f"Variable {variable_name} not found")



"""Verification metrics for forecasts."""
from sheerwater.utils import start_remote
from .metrics_tables import summary_metrics_table


if __name__ == "__main__":
    """Temporary runners to generate rainy onset tables."""
    start_remote(remote_name='genevieve', remote_config='xlarge_cluster')
    # Runners to generate the tables
    start_time = '2016-01-01'
    end_time = '2022-12-31'

    # variables = ['precip', 'tmp2m']
    # variables = ['rainy_onset_no_drought', 'rainy_onset']
    variables = ['rainy_onset_no_drought']
    # metrics = ['mae', 'rmse', 'bias', 'acc', 'heidke', 'pod', 'far', 'ets', 'mape', 'smape', 'bias_score', 'seeps']
    metrics = ['mae', 'bias', 'rmse']
    truth = 'era5'
    # time_grouping = [None, 'month', 'year']
    time_grouping = [None]
    region = 'kenya'
    grid = 'global1_5'
    mask = 'lsm'

    for variable in variables:
        for metric in metrics:
            for tg in time_grouping:
                ds = summary_metrics_table(start_time, end_time, variable, truth, metric,
                                           time_grouping=tg, grid=grid, mask=mask, region=region,
                                           recompute=True, cache_mode='overwrite',
                                           backend='sql')



"""Library of statistics implementations for verification."""
# flake8: noqa: D102, ARG001, D103

from functools import wraps

import numpy as np
import properscoring
import xarray as xr
from nuthatch import cache as cache_decorator
from nuthatch.processors import timeseries as timeseries_decorator
from sheerwater.interfaces import spatial

# Global metric registry dictionary
SHEERWATER_STATISTIC_REGISTRY = {}


def statistic(cache=False, name=None,
              timeseries='time',
              cache_args=['variable', 'agg_days', 'forecast', 'truth',
                          'data_key', 'grid', 'statistic'],
              chunking={"lat": 121, "lon": 240, "time": 30, 'region': 300, 'prediction_timedelta': -1},
              chunk_by_arg={
                  'grid': {
                      'global0_25': {"lat": 721, "lon": 1440, "time": 30}
                  },
              }):
    """A unifed statistics decorator.

    With this decorator, you can define a statistic function that will be cached and used to compute the statistic
    for a given forecast and observation. The statistic function will be provided with the forecast, observation, and
    data as arguments, and should return a xarray.DataArray.

    The statistic function will be cached using the cache decorator.

    """
    def create_statistic(func):
        # Create a global statistic as a cachable function with keys that accepts data
        @timeseries_decorator(timeseries=timeseries)
        @spatial()
        @cache_decorator(cache=cache, cache_args=cache_args,
                         backend_kwargs={
                             'chunking': chunking,
                             'chunk_by_arg': chunk_by_arg
                         })
        def global_statistic(
            data, data_key,
            start_time, end_time,
            variable, agg_days, forecast, truth,
            statistic, grid,
            mask=None, region='global',
            **cache_kwargs
        ):
            # Pass the cache kwargs through to the statistics function
            cache_kwargs = {
                'data_key': data_key,
                'start_time': start_time, 'end_time': end_time,
                'variable': variable, 'agg_days': agg_days, 'forecast': forecast, 'truth': truth,
                'grid': grid, 'mask': mask, 'region': region,
                'statistic': statistic,
            }
            ds = func(data=data, **cache_kwargs)
            # Assign metric attributes in one call
            ds = ds.assign_attrs(
                prob_type=data['prob_type'],
                forecast=forecast,
                truth=truth,
                statistic=statistic
            )
            return ds

        @wraps(func)
        def wrapper(
            data, **cache_kwargs
        ):
            # Set the statistic to the function name in lowercase
            cache_kwargs['statistic'] = name

            # Call the global statistic function
            ds = global_statistic(
                data=data, memoize=True,
                **cache_kwargs,
            )
            return ds

        # Register the wrapped function with the registry
        SHEERWATER_STATISTIC_REGISTRY[name] = wrapper

        return wrapper
    return create_statistic


@statistic(cache=False, name='obs')
def fn_obs(data, **cache_kwargs):  # noqa: F821
    return data['obs']


@statistic(cache=False, name='fcst')
def fn_fcst(data, **cache_kwargs):  # noqa: F821
    return data['fcst']


@statistic(cache=False, name='squared_obs')
def fn_squared_obs(data, **cache_kwargs):  # noqa: F821
    return data['obs']**2


@statistic(cache=False, name='squared_fcst')
def fn_squared_fcst(data, **cache_kwargs):  # noqa: F821
    return data['fcst']**2


@statistic(cache=False, name='covariance')
def fn_covariance(data, **cache_kwargs):  # noqa: F821
    return data['fcst'] * data['obs']


@statistic(cache=False, name='obs_anom')
def fn_obs_anom(data, **cache_kwargs):  # noqa: F821
    return data['obs'] - data['climatology']


@statistic(cache=False, name='fcst_anom')
def fn_fcst_anom(data, **cache_kwargs):  # noqa: F821
    return data['fcst'] - data['climatology']


@statistic(cache=False, name='squared_obs_anom')
def fn_squared_obs_anom(data, **cache_kwargs):  # noqa: F821
    obs_anom = fn_obs_anom(data, **cache_kwargs)
    return obs_anom**2


@statistic(cache=False, name='squared_fcst_anom')
def fn_squared_fcst_anom(data, **cache_kwargs):  # noqa: F821
    fcst_anom = fn_fcst_anom(data, **cache_kwargs)
    return fcst_anom**2


@statistic(cache=False, name='anom_covariance')
def fn_anom_covariance(data, **cache_kwargs):  # noqa: F821
    fcst_anom = fn_fcst_anom(data, **cache_kwargs)
    obs_anom = fn_obs_anom(data, **cache_kwargs)
    return obs_anom * fcst_anom


@statistic(cache=False, name='n_valid')
def fn_n_valid(data, **cache_kwargs):  # noqa: F821
    return xr.ones_like(data['fcst']).where(data['fcst'].notnull(), 0.0, drop=False).astype(float)


@statistic(cache=False, name='obs_digitized')
def fn_obs_digitized(data, **cache_kwargs):  # noqa: F821
    # `np.digitize(x, bins, right=True)` returns index `i` such that:
    #   `bins[i-1] < x <= bins[i]`
    # Indices range from 0 (for x <= bins[0]) to len(bins) (for x > bins[-1]).
    # `bins` for np.digitize will be `thresholds_np`.
    ds = xr.apply_ufunc(
        np.digitize,
        data['obs'],
        kwargs={'bins': data['bins'], 'right': True},
        dask='parallelized',
        output_dtypes=[int],
    )
    # Restore NaN values
    return ds.where(data['obs'].notnull(), np.nan, drop=False).astype(float)


@statistic(cache=False, name='fcst_digitized')
def fn_fcst_digitized(data, **cache_kwargs):  # noqa: F821
    # `np.digitize(x, bins, right=True)` returns index `i` such that:
    #   `bins[i-1] < x <= bins[i]`
    # Indices range from 0 (for x <= bins[0]) to len(bins) (for x > bins[-1]).
    # `bins` for np.digitize will be `thresholds_np`.
    ds = xr.apply_ufunc(
        np.digitize,
        data['fcst'],
        kwargs={'bins': data['bins'], 'right': True},
        dask='parallelized',
        output_dtypes=[int],
    )
    # Restore NaN values
    return ds.where(data['fcst'].notnull(), np.nan)


@statistic(cache=False, name='false_positives')
def fn_false_positives(data, **cache_kwargs):  # noqa: F821
    # "Positive events" are 2 (above threshold); negative are 1
    obs_dig = fn_obs_digitized(data, **cache_kwargs)
    fcst_dig = fn_fcst_digitized(data, **cache_kwargs)
    return (obs_dig == 1) & (fcst_dig == 2)


@statistic(cache=False, name='false_negatives')
def fn_false_negatives(data, **cache_kwargs):  # noqa: F821
    obs_dig = fn_obs_digitized(data, **cache_kwargs)
    fcst_dig = fn_fcst_digitized(data, **cache_kwargs)
    return (obs_dig == 2) & (fcst_dig == 1)


@statistic(cache=False, name='true_positives')
def fn_true_positives(data, **cache_kwargs):  # noqa: F821
    obs_dig = fn_obs_digitized(data, **cache_kwargs)
    fcst_dig = fn_fcst_digitized(data, **cache_kwargs)
    return (obs_dig == 2) & (fcst_dig == 2)


@statistic(cache=False, name='true_negatives')
def fn_true_negatives(data, **cache_kwargs):  # noqa: F821
    obs_dig = fn_obs_digitized(data, **cache_kwargs)
    fcst_dig = fn_fcst_digitized(data, **cache_kwargs)
    return (obs_dig == 1) & (fcst_dig == 1)


@statistic(cache=False, name='n_correct')
def fn_n_correct(data, **cache_kwargs):  # noqa: F821
    obs_dig = fn_obs_digitized(data, **cache_kwargs)
    fcst_dig = fn_fcst_digitized(data, **cache_kwargs)
    return (obs_dig == fcst_dig)


# Dynamically generate functions for each category and bind it with the correct category
def make_fn_n_obs_bin(category):
    @statistic(cache=False, name=f'n_obs_bin_{category}')
    def fn_n_obs_bin(data, **cache_kwargs):
        obs_dig = fn_obs_digitized(data, **cache_kwargs)
        return (obs_dig == category)
    return fn_n_obs_bin


def make_fn_n_fcst_bin(category):
    @statistic(cache=False, name=f'n_fcst_bin_{category}')
    def fn_n_fcst_bin(data, **cache_kwargs):
        fcst_dig = fn_fcst_digitized(data, **cache_kwargs)
        return (fcst_dig == category)
    return fn_n_fcst_bin


# Generate 10 categorical obserations by default for the categorical metrics
for cat in range(1, 10):
    make_fn_n_obs_bin(cat)
    make_fn_n_fcst_bin(cat)


@statistic(cache=False, name='mae')
def fn_mae(data, **cache_kwargs):  # noqa: F821
    return abs(data['fcst'] - data['obs'])


@statistic(cache=False, name='mse')
def fn_mse(data, **cache_kwargs):  # noqa: F821
    return (data['fcst'] - data['obs'])**2


@statistic(cache=False, name='bias')
def fn_bias(data, **cache_kwargs):  # noqa: F821
    return data['fcst'] - data['obs']


@statistic(cache=False, name='mape')
def fn_mape(data, **cache_kwargs):  # noqa: F821
    return abs(data['fcst'] - data['obs']) / np.maximum(abs(data['obs']), 1e-10)


@statistic(cache=False, name='smape')
def fn_smape(data, **cache_kwargs):  # noqa: F821
    return abs(data['fcst'] - data['obs']) / (abs(data['fcst']) + abs(data['obs']))


@statistic(cache=False, name='brier')
def fn_brier(data, **cache_kwargs):  # noqa: F821
    fcst_dig = fn_fcst_digitized(data, **cache_kwargs)
    obs_dig = fn_obs_digitized(data, **cache_kwargs)
    positive_event_id = 2
    fcst_event_prob = (fcst_dig == positive_event_id).mean(dim='member')
    obs_event_prob = (obs_dig == positive_event_id)
    return (fcst_event_prob - obs_event_prob)**2


@statistic(cache=True, name='seeps')
def fn_seeps(data, **cache_kwargs):  # noqa: F821
    """Based on the implementation in WeatherBench2."""
    wet_threshold = data['wet_threshold']
    if 'quantile' in wet_threshold.coords:
        wet_threshold = wet_threshold.drop_vars('quantile')
    dry_fraction = data['dry_fraction']
    dry_threshold_mm = 0.25

    # Get the average dry fraction over the entire year
    p1 = dry_fraction.mean("dayofyear")['dry_fraction']
    min_p1 = 0.03  # the minimum allowed dry fraction
    max_p1 = 0.93  # the maximum allowed dry fraction

    # fcst and obs should have the same valid times at this point
    valid_times = data['obs'].time
    wet_threshold_by_time = wet_threshold.sel(dayofyear=valid_times.dt.dayofyear)

    def convert_precips_to_seeps_cat(da, wet_threshold_by_time, dry_threshold_mm):
        # Convert to dry rains
        dry = da < dry_threshold_mm
        light = (da > dry_threshold_mm) & (da < wet_threshold_by_time)
        heavy = (da >= wet_threshold_by_time)
        result = xr.concat(
            [dry, light, heavy],
            dim=xr.DataArray(["dry", "light", "heavy"], dims=["seeps_cat"]),
        )
        # Convert NaNs back to NaNs
        result = result.astype("int").where(da.notnull())
        return result

    fcst_cat = convert_precips_to_seeps_cat(
        data['fcst']['precip'], wet_threshold_by_time['wet_threshold'], dry_threshold_mm)\
        .rename({'seeps_cat': 'fcst_cat'})
    obs_cat = convert_precips_to_seeps_cat(
        data['obs']['precip'], wet_threshold_by_time['wet_threshold'], dry_threshold_mm)\
        .rename({'seeps_cat': 'obs_cat'})

    # Get the elementwise product of the forecast and observation categories
    out = fcst_cat * obs_cat

    # The scoring matrix, which assigns a penalty for each of the 3 x 3 = 9
    # different category combinations
    scoring_matrix = [
        [xr.zeros_like(p1), 1 / (1 - p1), 4 / (1 - p1)],
        [1 / p1, xr.zeros_like(p1), 3 / (1 - p1)],
        [
            1 / p1 + 3 / (2 + p1),
            3 / (2 + p1),
            xr.zeros_like(p1),
        ],
    ]
    das = []
    for mat in scoring_matrix:
        das.append(xr.concat(mat, dim=out.obs_cat))
    scoring_matrix = 0.5 * xr.concat(das, dim=out.fcst_cat)

    # Take the dot product to compute the final scoring
    m_ds = xr.dot(out, scoring_matrix, dims=['fcst_cat', 'obs_cat'])

    # Mask out the p1 threshholds
    m_ds = m_ds.where(p1 < max_p1, np.nan)
    m_ds = m_ds.where(p1 > min_p1, np.nan)
    return xr.Dataset({'precip': m_ds})


@statistic(cache=False, name='crps')
def fn_crps(data, **cache_kwargs):  # noqa: F821
    if data['prob_type'] == 'ensemble':
        fcst = data['fcst'].chunk(member=-1, time=1, prediction_timedelta=1,
                                  lat=250, lon=250)  # member must be -1 to succeed
        m_ds = xr.apply_ufunc(
            properscoring.crps_ensemble,
            data['obs'],
            fcst,
            input_core_dims=[[], ['member']],
            kwargs={"axis": -1, "issorted": False, "weights": None},
            dask="parallelized",
            output_dtypes=[float],
            keep_attrs=True,
        )
    elif data['prob_type'] == 'quantile':
        # Custom implementation of a quantile CRPS approximation
        diff = data['fcst'] - data['obs']
        ind = diff > 0
        qnt_val = diff['member']
        qnt_score = 2 * (ind - qnt_val) * diff
        m_ds = qnt_score.integrate(coord='member')
    else:
        raise ValueError(f"Invalid probability type: {data['prob_type']}")
    return m_ds


def statistic_factory(statistic_name: str):
    """Get a statistic function by name from the registry."""
    try:
        return SHEERWATER_STATISTIC_REGISTRY[statistic_name.lower()]
    except KeyError:
        raise ValueError(f"Unknown statistic: {statistic_name}. Available statistics: {list_statistics()}")


def list_statistics():
    """List all available statistics in the registry."""
    return list(SHEERWATER_STATISTIC_REGISTRY.keys())



"""Mask data objects."""
import os

import cdsapi
import numpy as np
import rioxarray  # noqa: F401 - needed to enable .rio attribute
import xarray as xr
from nuthatch import cache

from sheerwater.utils import cdsapi_secret, clip_region, get_grid, get_grid_ds, region_data, lon_base_change


@cache(cache_args=['grid'])
def land_sea_mask(grid="global1_5"):
    """Get the ECMWF global land sea mask for the given grid.

    Args:
        grid (str): The grid to fetch the data at.  Note that only
            the resolution of the specified grid is used.
    """
    _, _, grid_size, _ = get_grid(grid, base="base360")

    # Get data from the CDS API
    times = ['00:00']
    days = ['01']
    months = ['01']

    # Make sure the temp folder exists
    os.makedirs('./temp', exist_ok=True)
    path = "./temp/lsm.nc"

    # Create a file path in the temp folder
    url, key = cdsapi_secret()
    c = cdsapi.Client(url=url, key=key)
    c.retrieve('reanalysis-era5-single-levels',
               {
                   'product_type': 'reanalysis',
                   'variable': "land_sea_mask",
                   'year': 2022,
                   'month': months,
                   'day': days,
                   'time': times,
                   'format': 'netcdf',
                   'grid': [str(grid_size), str(grid_size)],
               },
               path)

    # Some transformations
    ds = xr.open_dataset(path)

    ds = ds.rename({'latitude': 'lat', 'longitude': 'lon', 'lsm': 'mask'})

    if 'valid_time' in ds.coords:
        time_var = 'valid_time'
    elif 'time' in ds.coords:
        time_var = 'time'
    else:
        raise ValueError("Could not find time variable in dataset.")
    ds = ds.sel(**{time_var: ds[time_var].values[0]})
    ds = ds.drop(time_var)

    if 'expver' in ds.coords:
        ds = ds.drop('expver')

    # Convert to our standard base 180 format
    ds = lon_base_change(ds, to_base="base180")

    # Sort and select a subset of the data
    ds = ds.sortby(['lon', 'lat'])  # CDS API returns lat data in descending order, breaks slicing

    ds = ds.compute()
    os.remove(path)
    return ds

@cache(cache_args=['mask', 'grid'], memoize=True)
def spatial_mask(mask, grid='global1_5', region='global'):
    """Get a mask dataset.

    Args:
        mask (str): The mask to apply. One of: 'lsm', None
            To get different land-sea masks, use 'lsm-<value>'. For example, 'lsm-0.5' will return a mask
            where the mask is greater than 0.5. Defaults to 0.0.
        grid (str): The grid resolution of the dataset.
        region (str): The region to clip to. A specific instance of a space group

    Returns:
        xr.Dataset: Mask dataset.
    """
    if mask is None:
        return get_grid_ds(grid)
    elif 'lsm' in mask:
        # Import here to avoid circular imports
        if grid == 'global1_5' or grid == 'global0_25':
            mask_ds = land_sea_mask(grid=grid).compute()
        else:
            # TODO: Should implement a more resolved land-sea mask for the other grids
            from sheerwater.utils.data_utils import regrid
            mask_ds = land_sea_mask(grid='global0_25')
            mask_ds = regrid(mask_ds, grid, method='nearest', region=region).compute()

        val = 0.0
        if '-' in mask:
            # Convert to boolean mask
            val = float(mask.split('-')[1])
        mask_ds['mask'] = mask_ds['mask'] > val
        return mask_ds
    else:
        raise NotImplementedError("Only land-sea or None mask is implemented.")



@cache(cache_args=['grid', 'space_grouping', 'region'],
       backend_kwargs={'chunking': {'lat': 1800, 'lon': 3600}})
def region_labels(grid='global1_5', space_grouping='country', region='global'):
    """Generate a dataset with a region coordinate at a specific space grouping.

    Available space groupings are
     - 'country', 'continent', 'subregion', 'region_un', 'region_wb', 'meteorological_zone',
        'hemisphere', 'global', and 'sheerwater_region'.

    Args:
        grid (str): The grid to fetch the data at.  Note that only
            the resolution of the specified grid is used.
        space_grouping (str):
            - country, continent, subregion, region_un, region_wb, meteorological_zone,
              hemisphere, sheerwater_region
        region (str): The region to clip to. A specific instance of a space group
            -global, or any specific instance of the space groupings above, e.g., africa


    Returns:
        xarray.Dataset: Dataset with added region coordinate
    """
    # Get the list of regions for the specified admin level
    region_df = region_data(space_grouping)

    # Get the grid dataframe
    ds = get_grid_ds(grid)
    # Assign a dummy region coordinate to all grid cells
    # Fixed data type of strings of length 40
    ds = ds.assign_coords(region=(('lat', 'lon'), xr.full_like(ds.lat * ds.lon, 'no_region', dtype='U40').data))
    if region != 'global':
        ds = clip_region(ds, region=region)

    # Loop through each region and label grid cells
    for i, rn in region_df.iterrows():
        print(i+1, '/', len(region_df.region_name), rn.region_name)
        # Clip the grid to the boundary of Shapefile
        world_ds = xr.full_like(ds.lat * ds.lon, 1.0, dtype=np.float32)
        #  Add geometry to the dataframe and clip
        world_ds = world_ds.rio.write_crs("EPSG:4326")
        world_ds = world_ds.rio.set_spatial_dims('lon', 'lat')
        region_ds = world_ds.rio.clip(rn, region_df.crs, drop=False)
        # Assign the region name to the region coordinate
        ds['region'] = xr.where(~region_ds.isnull(), rn.region_name, ds['region'])

    return ds


__all__ = ['land_sea_mask', 'region_labels']



"""Verification metrics for forecasters and reanalyses."""
import xarray as xr
from nuthatch import cache

from sheerwater.metrics_library import metric_factory
from sheerwater.interfaces import get_data
from sheerwater.regions_and_masks import region_labels, spatial_mask
from sheerwater.utils import dask_remote, groupby_region, groupby_time, clip_region


@dask_remote
@cache(cache_args=['start_time', 'end_time', 'variable', 'agg_days', 'forecast', 'truth',
                   'metric_name', 'time_grouping', 'space_grouping', 'spatial', 'grid', 'mask', 'region'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "time": 100, 'region': 300, 'prediction_timedelta': -1},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, "time": 30}
               },
           }
})
def metric(start_time, end_time, variable, agg_days, forecast, truth,
           metric_name, time_grouping=None, space_grouping=None,
           spatial=False, grid="global1_5", mask='lsm', region='global'):
    """Compute a grouped metric for a forecast at a specific lead."""
    # Use the metric registry to get the metric class
    metric_obj = metric_factory(metric_name, start_time=start_time, end_time=end_time, variable=variable,
                                agg_days=agg_days, forecast=forecast, truth=truth, time_grouping=time_grouping,
                                space_grouping=space_grouping, spatial=spatial, grid=grid, mask=mask, region=region)
    return metric_obj.compute()


@dask_remote
@cache(cache_args=['start_time', 'end_time', 'variable', 'agg_days', 'station_data',
                   'time_grouping', 'grid', 'mask', 'region', 'missing_thresh'])
def coverage(start_time=None, end_time=None, variable='precip', agg_days=7, station_data='ghcn_avg',
             time_grouping=None, space_grouping=None, grid="global1_5", mask='lsm',
             region='global', missing_thresh=0.9):  # noqa: ARG001
    """Compute coverage of a dataset."""
    # Use the metric registry to get the metric class
    station_data_fn = get_data(station_data)
    data = station_data_fn(start_time, end_time, variable, agg_days=agg_days,
                           grid=grid, mask=None, region=region)

    data['non_null'] = data[variable].notnull()
    data['indicator'] = xr.ones_like(data[variable])
    data = groupby_time(data, time_grouping=time_grouping, agg_fn='mean')

    region_ds = region_labels(grid=grid, space_grouping=space_grouping, region=region).compute()
    mask_ds = spatial_mask(mask=mask, grid=grid, memoize=True)
    if region != 'global':
        mask_ds = clip_region(mask_ds, region=region)
    data = groupby_region(data, region_ds, mask_ds, agg_fn='sum')

    data['coverage'] = data['non_null'] / data['indicator']

    data = data.drop_vars(variable)
    return data


__all__ = ['metric']



"""A climatology baseline forecast for benchmarking."""
from datetime import datetime

import dask
import dateparser
import numpy as np
import pandas as pd
import xarray as xr
from dateutil.relativedelta import relativedelta
from nuthatch import cache
from nuthatch.processors import timeseries

from sheerwater.interfaces import forecast as sheerwater_forecast, spatial
from sheerwater.reanalysis import era5_daily, era5_rolled
from sheerwater.utils import add_dayofyear, dask_remote, get_dates, pad_with_leapdays


@dask_remote
@spatial()
@cache(cache_args=['first_year', 'last_year', 'agg_days', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 721, "lon": 1440, "dayofyear": 30}
})
def seeps_dry_fraction(first_year=1985, last_year=2014, agg_days=7, grid='global1_5', mask=None, region='global'):
    """Compute the climatology of the ERA5 data. Years are inclusive."""
    start_time = f"{first_year}-01-01"
    end_time = f"{last_year}-12-31"

    # Get the rolled era5 data
    ds = era5_rolled(start_time, end_time, variable='precip', agg_days=agg_days, grid=grid, mask=mask, region=region)

    # Add day of year as a coordinate
    ds = add_dayofyear(ds)
    ds = pad_with_leapdays(ds)

    ds['is_dry'] = (ds['precip'] < 0.25)
    ds = ds.groupby('dayofyear').mean(dim='time')
    ds = ds.drop_vars(['precip'])
    ds = ds.rename({
        'is_dry': 'dry_fraction',
    })

    # Convert to true day of year
    ds['dayofyear'] = ds.dayofyear.dt.dayofyear

    return ds


@dask_remote
@spatial()
@cache(cache_args=['first_year', 'last_year', 'agg_days', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 721, "lon": 1440, "dayofyear": 30}
})
def seeps_wet_threshold(first_year=1985, last_year=2014, agg_days=7, grid='global1_5', mask=None, region='global'):
    """Compute the climatology of the ERA5 data. Years are inclusive."""
    start_time = f"{first_year}-01-01"
    end_time = f"{last_year}-12-31"

    # Get the rolled era5 data
    ds = era5_rolled(start_time, end_time, variable='precip', agg_days=agg_days, grid=grid, mask=mask, region=region)

    # Add day of year as a coordinate
    ds = add_dayofyear(ds)
    ds = pad_with_leapdays(ds)

    ds = ds.groupby('dayofyear').quantile(2/3, method='nearest', dim='time')

    ds = ds.rename({
        'precip': 'wet_threshold',
    })

    # Convert to true day of year
    ds['dayofyear'] = ds.dayofyear.dt.dayofyear

    return ds


@dask_remote
@spatial()
@cache(cache_args=['variable', 'first_year', 'last_year', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 721, "lon": 1440, "dayofyear": 30}
})
def climatology_raw(variable, first_year=1985, last_year=2014, grid='global1_5', mask=None, region='global'):
    """Compute the climatology of the ERA5 data. Years are inclusive."""
    start_time = f"{first_year}-01-01"
    end_time = f"{last_year}-12-31"

    # Get single day, masked data between start and end years
    ds = era5_daily(start_time, end_time, variable=variable, grid=grid, mask=mask, region=region)

    # Add day of year as a coordinate
    ds = add_dayofyear(ds)
    ds = pad_with_leapdays(ds)

    # Take average over the period to produce climatology that includes leap years
    ds = ds.groupby('dayofyear').mean(dim='time')

    return ds


@dask_remote
@spatial()
@cache(cache_args=['variable', 'first_year', 'last_year', 'prob_type', 'agg_days', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "dayofyear": 1000, "member": 1},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'dayofyear': 30, 'member': 1}
               }
           }
})
def climatology_agg_raw(variable, first_year=1985, last_year=2014,
                        prob_type='deterministic', agg_days=7, grid="global1_5", mask=None, region='global'):
    """Generates aggregated climatology."""
    start_time = f"{first_year}-01-01"
    end_time = f"{last_year}-12-31"
    ds = era5_rolled(start_time, end_time, variable=variable, agg_days=agg_days, grid=grid, mask=mask, region=region)

    # Add day of year as a coordinate
    ds = add_dayofyear(ds)
    ds = pad_with_leapdays(ds)

    # Take average over the period to produce climatology
    if prob_type == 'deterministic':
        return ds.groupby('dayofyear').mean(dim='time')
    elif prob_type == 'probabilistic':
        # Otherwise, get ensemble members sampled from climatology
        def sample_members(sub_ds, members=30):
            doy = sub_ds.dayofyear.values[0]
            ind = np.random.randint(0, len(sub_ds.time.values), size=(members,))
            sub = sub_ds.isel(time=ind)
            sub = sub.assign_coords(time=np.arange(members)).rename({'time': 'member'})
            sub = sub.assign_coords(dayofyear=doy)
            return sub

        doys = []
        for doy in np.unique(ds.dayofyear.values):
            doys.append(
                sample_members(ds.isel(time=(ds.dayofyear.values == doy))))
        ds = xr.concat(doys, dim='dayofyear')
        ds = ds.chunk(member=1)
        return ds
    else:
        raise ValueError(f"Unsupported prob_type: {prob_type}")


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'clim_years', 'agg_days', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "time": 1000},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'time': 30}
               }
           }
})
def climatology_rolling_agg(start_time, end_time, variable, clim_years=30,
                            agg_days=7, grid="global1_5", mask=None,
                            region='global'):
    """Compute a rolling {clim_years}-yr climatology of the ERA5 data.

    Args:
        start_time: First time of the forecast period.
        end_time: Last time of the forecast period.
        variable: Variable to compute climatology for.
        clim_years: Number of years to compute climatology over.
        agg_days (int): Aggregation period in days.
        grid: Grid resolution of the data.
        mask: Spatial mask to apply.
        region: Region to compute climatology for.
    """
    #  Get reanalysis data for the appropriate look back period
    # We need data from clim_years before the start_time until 1 year before the end_time
    # as this climatology excludes the most recent year for use in operational forecasting
    new_start = (dateparser.parse(start_time) - relativedelta(years=clim_years)).strftime("%Y-%m-%d")
    new_end = (dateparser.parse(end_time) - relativedelta(years=1)).strftime("%Y-%m-%d")

    # Get ERA5 data, and ignore cache validation if start_time is earlier than the cache
    ds = era5_rolled(new_start, new_end, variable=variable, agg_days=agg_days, grid=grid, mask=mask, region=region)
    ds = add_dayofyear(ds)
    ds = pad_with_leapdays(ds)

    def doy_rolling(sub_ds, years):
        return sub_ds.rolling(time=years, min_periods=years, center=False).mean()

    # Rechunk the data to have a single time chunk for efficient rolling
    ds = ds.chunk(time=1)
    ds = ds.groupby('dayofyear').map(doy_rolling, years=clim_years)
    ds = ds.dropna('time', how='all')
    ds = ds.drop('dayofyear')
    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'agg_days', 'grid'],
       backend_kwargs={'chunking': {"lat": 300, "lon": 300, "time": 366}})
def _era5_rolled_for_clim(start_time, end_time, variable, agg_days=7,
                          grid="global1_5", mask=None, region='global'):
    """Aggregates the hourly ERA5 data into daily data and rolls.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        agg_days (int): The aggregation period to use, in days
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
            - global0_25: 0.25 degree global grid
        mask: Spatial mask to apply.
        region: Region to fetch data for.
    """
    # Get single day, masked data between start and end years
    ds = era5_rolled(start_time, end_time, variable=variable, agg_days=agg_days, grid=grid, mask=mask, region=region)

    # Add day of year as a coordinate
    ds = add_dayofyear(ds)
    ds = pad_with_leapdays(ds)
    ds = ds.assign_coords(year=ds.time.dt.year)
    ds = ds.chunk({'lat': 300, 'lon': 300, 'time': 366})
    return ds


@dask_remote
@spatial()
@cache(cache_args=['variable', 'first_year', 'last_year', 'agg_days', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "dayofyear": 366},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'dayofyear': 30}
               }
           }
})
def climatology_linear_weights(variable, first_year=1985, last_year=2014,
                               agg_days=7, grid='global1_5', mask=None,
                               region='global'):
    """Fit the climatological trend for a specific day of year.

    Args:
        variable: Variable to compute climatology for.
        first_year: First year of the climatology.
        last_year: Last year of the climatology.
        agg_days: Aggregation period in days.
        grid: Grid resolution of the data.
        mask: Spatial mask to apply.
        region: Region to compute climatology for.
    """
    start_time = f"{first_year}-01-01"
    end_time = f"{last_year}-12-31"

    # Get single day, masked data between start and end years
    ds = _era5_rolled_for_clim(start_time, end_time, variable=variable,
                               agg_days=agg_days, grid=grid, mask=mask,
                               region=region)

    def fit_trend(sub_ds):
        return sub_ds.swap_dims({"time": "year"}).polyfit(dim='year', deg=1)
    # Fit the trend for each day of the year
    ds = ds.groupby('dayofyear').map(fit_trend)
    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache=False,
       cache_args=['variable', 'first_year', 'last_year', 'trend', 'prob_type', 'agg_days', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "time": 1000},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'time': 30}
               }
           }
       })
def climatology_rolled(start_time, end_time, variable, first_year=1985,
                       last_year=2014, trend=False, prob_type='deterministic',
                       agg_days=7, grid="global1_5", mask=None,
                       region='global'):
    """Generates a forecast timeseries of climatology.

    Args:
        start_time (str): The start time of the timeseries forecast.
        end_time (str): The end time of the timeseries forecast.
        variable (str): The weather variable to fetch.
        first_year (int): The first year to use for the climatology.
        last_year (int): The last year to use for the climatology.
        trend (bool): Whether to include a trend in the forecast.
        prob_type (str): The type of forecast to generate.
        agg_days (int): The aggregation period to use, in days
        grid (str): The grid to produce the forecast on.
        mask: Spatial mask to apply.
        region: Region to compute climatology for.
    """
    # Create a target date dataset
    target_dates = get_dates(start_time, end_time, stride='day', return_string=False)
    time_ds = xr.Dataset({'time': target_dates})
    time_ds = add_dayofyear(time_ds)

    if trend:
        if prob_type == 'probabilistic':
            raise NotImplementedError("Probabilistic trend forecasts are not supported.")

        time_ds = time_ds.assign_coords(year=time_ds['time'].dt.year)
        coeff = climatology_linear_weights(variable, first_year=first_year, last_year=last_year,
                                           agg_days=agg_days, grid=grid, mask=mask, region=region)
        with dask.config.set(**{'array.slicing.split_large_chunks': True}):
            coeff = coeff.sel(dayofyear=time_ds.dayofyear)
            coeff = coeff.drop('dayofyear')

        def linear_fit(coeff):
            """Compute the linear fit y = a * year + b for the given coefficients."""
            est = coeff[f"{variable}_polyfit_coefficients"].sel(degree=1) * coeff["year"].astype("float") \
                + coeff[f"{variable}_polyfit_coefficients"].sel(degree=0)
            est = est.to_dataset(name=variable)
            if variable == 'precip':
                est = np.maximum(est, 0)
            return est
        ds = linear_fit(coeff)
        ds = ds.drop('year')
    else:
        # Get climatology on the corresponding global grid
        ds = climatology_agg_raw(variable, first_year=first_year, last_year=last_year,
                                 prob_type=prob_type, agg_days=agg_days, grid=grid, mask=mask, region=region)
        # Select the climatology data for the target dates, and split large chunks
        with dask.config.set(**{'array.slicing.split_large_chunks': True}):
            ds = ds.sel(dayofyear=time_ds.dayofyear)
            ds = ds.drop('dayofyear')
    return ds


@dask_remote
def _climatology_unified(start_time, end_time, variable, agg_days,
                         first_year=1985, last_year=2014, trend=False,
                         prob_type='deterministic', grid='global0_25', mask=None, region='global'):
    """Standard format forecast data for climatology forecast."""
    ds = climatology_rolled(start_time, end_time, variable,
                            first_year=first_year, last_year=last_year,
                            trend=trend,
                            prob_type=prob_type,
                            agg_days=agg_days,
                            grid=grid, mask=mask, region=region)

    if prob_type == 'deterministic':
        ds = ds.assign_attrs(prob_type="deterministic")
    else:
        ds = ds.assign_attrs(prob_type="ensemble")

    # To match the standard forecast format, add a prediction_timedelta coordinate
    ds = ds.expand_dims({"prediction_timedelta": [np.timedelta64(0, "ns")]})  # nanosecond precision
    ds = ds.rename({"time": "init_time"})
    return ds


@dask_remote
@sheerwater_forecast()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'prob_type', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365, 'lead_time': 1, 'member': 1}})
def climatology_2015(start_time, end_time, variable, agg_days=7, prob_type='deterministic',
                     grid='global0_25', mask='lsm', region='global'):
    """Standard format forecast data for climatology forecast."""
    return _climatology_unified(start_time, end_time, variable, agg_days=agg_days, first_year=1985, last_year=2014,
                                trend=False, prob_type=prob_type, grid=grid, mask=mask, region=region)


@dask_remote
@sheerwater_forecast()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'prob_type', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365, 'lead_time': 1, 'member': 1}})
def climatology_2020(start_time, end_time, variable, agg_days=7, prob_type='deterministic',
                     grid='global0_25', mask='lsm', region='global'):
    """Standard format forecast data for climatology forecast."""
    return _climatology_unified(start_time, end_time, variable, agg_days=agg_days, first_year=1990, last_year=2019,
                                trend=False, prob_type=prob_type, grid=grid, mask=mask, region=region)


@dask_remote
@sheerwater_forecast()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'prob_type', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365, 'lead_time': 1, 'member': 1}})
def climatology_trend_2015(start_time, end_time, variable, agg_days, prob_type='deterministic',
                           grid='global0_25', mask='lsm', region='global'):
    """Standard format forecast data for climatology forecast."""
    return _climatology_unified(start_time, end_time, variable, agg_days=agg_days, first_year=1985, last_year=2014,
                                trend=True, prob_type=prob_type, grid=grid, mask=mask, region=region)


@dask_remote
@sheerwater_forecast()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'prob_type', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365, 'lead_time': 1, 'member': 1}})
def climatology_rolling(start_time, end_time, variable, agg_days, prob_type='deterministic',
                        grid='global0_25', mask='lsm', region='global'):
    """Standard format forecast data for climatology forecast."""
    if prob_type != 'deterministic':
        raise NotImplementedError("Only deterministic forecasts are available for rolling climatology.")

    # Get daily data
    start_dt = dateparser.parse(start_time)
    start_dt -= relativedelta(years=1)  # exclude the most recent year for operational forecasting (handles leap year)
    new_start = datetime.strftime(start_dt, "%Y-%m-%d")

    end_dt = dateparser.parse(end_time)
    end_dt -= relativedelta(years=1)  # exclude the most recent year for operational forecasting (handles leap year)
    new_end = datetime.strftime(end_dt, "%Y-%m-%d")

    ds = climatology_rolling_agg(new_start, new_end, variable,
                                 clim_years=30, agg_days=agg_days,
                                 grid=grid, mask=mask, region=region)

    # Undo yearly time shifting
    times = [x + pd.DateOffset(years=1) for x in ds.time.values]
    ds = ds.assign_coords(time=times)

    # TODO: need to think through the padding with leap days, as we're getting duplicates
    ds = ds.drop_duplicates('time')
    # To match the standard forecast format, add a prediction_timedelta coordinate
    ds = ds.expand_dims({"prediction_timedelta": [np.timedelta64(0, "ns")]})
    ds = ds.rename({"time": "init_time"})
    return ds



"""A decorator for identifying data sources."""

import xarray as xr

from nuthatch.processor import NuthatchProcessor
from sheerwater.utils import convert_init_time_to_pred_time, clip_region, apply_mask

import logging
logger = logging.getLogger(__name__)

# Global registry of data sources
DATA_REGISTRY = {}
FORECAST_REGISTRY = {}

class SheerwaterDataset(NuthatchProcessor):
    """Processor for a Sheerwater dataset, either forecast or data of a standard format.

    This processor is used to convert forecast with init time and prediction timedelta to a valid time coordinate.

    It also implements spatial clipping and masking, and timeseries filtering.

    It supports xarray datasets.
    """

    def __init__(self, region_dim=None, **kwargs):
        """Initialize the spatial processor.

        Args:
            region_dim (str): The name of the region dimension. If None, the returned dataset will not be
                assumed to have a region dimesion and region data will be fetched from the region registry
                before clipping.
            kwargs: Additional keyword arguments to pass to the NuthatchProcessor.
        """
        super().__init__(**kwargs)
        self.region_dim = region_dim

    def process_arguments(self, sig, *args, **kwargs):
        """Process the arguments for the datasets decorator."""
        # Get default values for the function signature
        bound_args = self.bind_signature(sig, *args, **kwargs)

        # Default to global region and no masking if not passed
        self.region = bound_args.arguments.get('region', 'global')
        self.mask = bound_args.arguments.get('mask', None)
        try:
            self.grid = bound_args.arguments['grid']
            self.agg_days = bound_args.arguments['agg_days']
            self.variable = bound_args.arguments['variable']
        except KeyError:
            raise ValueError("Forecast decorator requires grid, agg_days, and variable to be passed.")

        if self.variable == 'precip':
            self.units = 'mm / day'
        elif self.variable == 'tmp2m':
            self.units = 'avg. daily C'
        else:
            self.units = None
        return args, kwargs

    def post_process(self, ds):
        """Post-process the dataset to implement masking and region clipping and timeseries postprocessing."""
        if isinstance(ds, xr.Dataset):
            # Clip to specified region
            if not (hasattr(ds, 'region') and ds.region== self.region):
                # Only clip region if the dataframe hasn't already been clipped
                ds = clip_region(ds, region=self.region, region_dim=self.region_dim)
            if not (hasattr(ds, 'mask') and ds.mask == self.mask):
                # Only apply mask if this dataframe has not already been masked
                ds = apply_mask(ds, self.mask, grid=self.grid)
            attrs = {
                'agg_days': float(self.agg_days),
                'variable': self.variable,
                'grid': self.grid,
                'mask': self.mask,
                'region': self.region
            }
            if self.units is not None:
                attrs['units'] = self.units
            ds = ds.assign_attrs(attrs)

        else:
            raise RuntimeError(f"Cannot clip by region and mask for data type {type(ds)}")
        return ds

    def validate(self, ds):
        """Validate the cached data to ensure it has data within the region."""
        # Check to see if the dataset extends roughly the full time series set
        test = clip_region(ds, region=self.region, region_dim=self.region_dim)
        test = apply_mask(test, self.mask, grid=self.grid)
        if test.notnull().count().compute() == 0:
            logger.warning(f"""The cached array does not have data within
                        the region {self.region}. Triggering recompute.
                        If you do not want to recompute the result set
                        `validate_data=False`""")
            return False
        else:
            return True

class data(SheerwaterDataset):
    """Processor for a Sheerwater data. It supports xarray datasets."""
    def __call__(self, func):
        """Call the parent class and register the data in the global data registry."""
        wrapped = super().__call__(func)
        DATA_REGISTRY[func.__name__] = wrapped
        return wrapped

    def post_process(self, ds):
        """Post-processor for a Sheerwater data. It supports xarray datasets."""
        # Implment masking and region clipping and timeseries postprocessing
        ds = super().post_process(ds)

        # Remove all unneeded dimensions
        ds = ds.drop_vars([var for var in ds.coords if var not in [
                            'time', 'lat', 'lon']])
        return ds

class forecast(SheerwaterDataset):
    """Processor for a Sheerwater forecast. It supports xarray datasets."""
    def __call__(self, func):
        """Call the forecast decorator and register it in the global forecast registry."""
        wrapped = super().__call__(func)
        FORECAST_REGISTRY[func.__name__] = wrapped
        return wrapped

    def post_process(self, ds):
        """Convert the init time and prediction timedelta to a valid time coordinate labeled 'time'."""
        # # Convert the init time and prediction timedelta to a valid time coordinate labeled 'time'
        if 'init_time' in ds.coords and 'prediction_timedelta' in ds.coords:
            # If we haven't converted from init time to valid times, do so now
            ds = convert_init_time_to_pred_time(ds)

        # Implment masking and region clipping and timeseries postprocessing
        ds = super().post_process(ds)

        # Remove all unneeded dimensions
        ds = ds.drop_vars([var for var in ds.coords if var not in [
                            'time', 'prediction_timedelta', 'lat', 'lon', 'member']])
        return ds

def get_forecast(forecast_name):
    """Get a forecast from the global forecast registry."""
    # The imports below ensure that all forecast modules are registered when this is called.
    import sheerwater.forecasts  # noqa: F401
    import sheerwater.climatology  # noqa: F401
    return FORECAST_REGISTRY[forecast_name]


def list_forecasts():
    """List all forecasts in the global forecast registry."""
    # The imports below ensure that all forecast modules are registered when this is called.
    import sheerwater.forecasts # noqa: F401
    import sheerwater.climatology # noqa: F401
    return list(FORECAST_REGISTRY.keys())


def get_data(data_name):
    """Get a data source from the global data registry."""
    # The imports below ensure that all forecast modules are registered when this is called.
    import sheerwater.data # noqa: F401
    import sheerwater.climatology # noqa: F401
    import sheerwater.reanalysis # noqa: F401
    return DATA_REGISTRY[data_name]


def list_data():
    """List all data sources in the global data registry."""
    # The imports below ensure that all forecast modules are registered when this is called.
    import sheerwater.data # noqa: F401
    import sheerwater.climatology # noqa: F401
    import sheerwater.reanalysis # noqa: F401
    return list(DATA_REGISTRY.keys())



"""Standard interfaces for Sheerwater data and forecasts."""
from .datasets import (forecast, data, DATA_REGISTRY, FORECAST_REGISTRY,  \
                        list_forecasts, get_forecast, list_data, get_data)
from .spatial import spatial

__all__ = ["forecast", "data", "spatial", "DATA_REGISTRY", "FORECAST_REGISTRY",
           "list_forecasts", "get_forecast", "list_data", "get_data"]



"""The task functions for the benchmarking platform."""
import numpy as np


def first_satisfied_date(ds, condition, time_dim='time', base_time=None,
                         prob_type='ensemble', prob_dim='member', prob_threshold=0.5):
    """Find the first date that a condition is satisfied in a timeseries.

    If the time dimension is a timedelta object, a base time must be specified.
    If a prob_dim is specified, find the first date that the condition is met with
    a probability greater than prob_threshold.

    Args:
        ds (xr.Dataset): Dataset to apply the condition to.
        condition (callable): Condition to apply to the dataset.
        time_dim (str): Name of the time dimension.
        base_time (str): Base time for timedelta objects (optional).
        prob_type (str): Type of probabilistic forecast. One of 'ensemble', 'quantile', or 'deterministic'.
        prob_dim (str): Name of the ensemble dimension.
        prob_threshold (float): Threshold for the probability dimension.
    """
    # Apply the rainy reason onset condition to the grouped dataframe
    ds['condition'] = condition(ds, prob_type=prob_type, prob_dim=prob_dim, prob_threshold=prob_threshold)

    # Ensure that dates are sorted
    ds = ds.sortby(time_dim)

    # Check if the time dimension is a timedelta object
    if np.issubdtype(ds[time_dim].dtype, np.timedelta64):
        if base_time is None:
            raise ValueError("If using timedelta64, must specify a base time for the timedelta.")
        fill_value = np.timedelta64('NaT')
    else:
        if base_time is not None:
            raise ValueError("Base time should only be specified if using a timedelta64 object.")
        fill_value = np.datetime64('NaT')

    # Get the first date that the condition is met; fill with known NaN value
    first_date = ds.condition.idxmax(dim=time_dim, fill_value=fill_value)
    # If the max value is the same as the first value and the first value is 0,
    # the condition was never met
    first_date = first_date.where((first_date > ds[time_dim].values[0]) |
                                  (ds.condition.isel({time_dim: 0}) == 1), other=fill_value)

    if np.issubdtype(ds[time_dim].dtype, np.timedelta64):
        # Add timedelta to the base time
        first_date = first_date[base_time] + first_date

    # Rename the variable
    first_date = first_date.rename('first_occurrence')
    return first_date



"""A decorator for spatial data."""
import xarray as xr

from nuthatch.processor import NuthatchProcessor

from sheerwater.utils import clip_region, apply_mask

import logging
logger = logging.getLogger(__name__)


class spatial(NuthatchProcessor):
    """Processor for spatial data.

    This processor is used to slice a spatial dataset based on the region.

    It also validates the spatial data to ensure it has data within the region.

    It supports xarray datasets.
    """

    def __init__(self, region_dim=None, **kwargs):
        """Initialize the spatial processor.

        Args:
            region_dim (str): The name of the region dimension. If None, the returned dataset will not be
                assumed to have a region dimesion and region data will be fetched from the region registry
                before clipping.
            kwargs: Additional keyword arguments to pass to the NuthatchProcessor.
        """
        super().__init__(**kwargs)
        self.region_dim = region_dim

    def process_arguments(self, sig, *args, **kwargs):
        """Process the arguments for the spatial decorator."""
        # Get default values for the function signature
        bound_args = self.bind_signature(sig, *args, **kwargs)

        # Default to global region and no masking if not passed
        self.region = bound_args.arguments.get('region', 'global')
        self.mask = bound_args.arguments.get('mask', None)
        try:
            self.grid = bound_args.arguments['grid']
        except KeyError:
            print("WARNING: No grid passed to spatial decorator. Cannot perform masking.")
            self.mask = None
            self.grid = None
        return args, kwargs

    def post_process(self, ds):
        """Post-process the dataset to clip to the region and apply the mask."""
        if isinstance(ds, xr.Dataset):
            # Clip to specified region
            if not (hasattr(ds, 'region') and ds.region== self.region):
                # Only clip region if the dataframe hasn't already been clipped
                ds = clip_region(ds, region=self.region, region_dim=self.region_dim)
            if not (hasattr(ds, 'mask') and ds.mask == self.mask):
                # Only apply mask if this dataframe has not already been masked
                ds = apply_mask(ds, self.mask, grid=self.grid)
            attrs = {
                'grid': self.grid,
                'mask': self.mask,
                'region': self.region
            }
            ds = ds.assign_attrs(attrs)
        else:
            raise RuntimeError(f"Cannot clip by region and mask for data type {type(ds)}")

        return ds

    def validate(self, ds):
        """Validate the cached data to ensure it has data within the region."""
        if isinstance(ds, xr.Dataset):
            # Check to see if the dataset extends roughly the full time series set
            test = clip_region(ds, region=self.region, region_dim=self.region_dim)
            test = apply_mask(test, self.mask, grid=self.grid)
            if test.notnull().count().compute() == 0:
                print("""WARNING: The cached array does not have data within
                          the region {self.region}. If you want to continue, set `validate_data=False`""")
                return False
            else:
                return True
        else:
            raise RuntimeError(f"Cannot validate region for data type {type(ds)}")
